from sklearn.svm import SVC as SkLeanSVC
from sklearn.svm import SVR as SkLearnSVR


class SVC(SkLeanSVC):
    """A Helix implementation of scikit-learn's SVC.

    It is exactly the same, except it always sets `probability` to `True`
    to get access to the `predict_proba` method, and `max_iter` to 1000, similar
    to `LinearSVC`, to avoid hanging on infinte iterations.
    """

    def __init__(
        self,
        *,
        C=1,
        kernel="rbf",
        degree=3,
        gamma="scale",
        coef0=0,
        shrinking=True,
        probability=True,  # set probability to True to access predict_proba
        tol=0.001,
        cache_size=200,
        class_weight=None,
        verbose=False,
        max_iter=1000,  # set max_iter to 1000 to avoid infinite fitting loop
        decision_function_shape="ovr",
        break_ties=False,
        random_state=None,
    ):
        super().__init__(
            C=C,
            kernel=kernel,
            degree=degree,
            gamma=gamma,
            coef0=coef0,
            shrinking=shrinking,
            probability=probability,
            tol=tol,
            cache_size=cache_size,
            class_weight=class_weight,
            verbose=verbose,
            max_iter=max_iter,
            decision_function_shape=decision_function_shape,
            break_ties=break_ties,
            random_state=random_state,
        )


class SVR(SkLearnSVR):
    """A Helix implementation of scikit-learn's SVR.

    It is exactly the same, except it always sets `max_iter` to 1000, similar
    to `LinearSVR`, to avoid hanging on infinte iterations.
    """

    def __init__(
        self,
        *,
        kernel="rbf",
        degree=3,
        gamma="scale",
        coef0=0,
        tol=0.001,
        C=1,
        epsilon=0.1,
        shrinking=True,
        cache_size=200,
        verbose=False,
        max_iter=1000,  # set max_iter to 1000 to avoid infinite fitting loop
    ):
        super().__init__(
            kernel=kernel,
            degree=degree,
            gamma=gamma,
            coef0=coef0,
            tol=tol,
            C=C,
            epsilon=epsilon,
            shrinking=shrinking,
            cache_size=cache_size,
            verbose=verbose,
            max_iter=max_iter,
        )



from dataclasses import dataclass


@dataclass
class MachineLearningOptions:
    model_types: dict
    save_actual_pred_plots: bool = True
    ml_log_dir: str = "ml"
    save_models: bool = True
    ml_plot_dir: str = "ml"
    use_hyperparam_search: bool = True


# ----- Bayesian Regularised Neural Network Parameters ----- >>>>>>>>


@dataclass
class BrnnOptions:
    """
    This class contains the parameters as an options
    for the Bayesian Regularised Neural Network.
    """

    batch_size: int = 32
    epochs: int = 10
    hidden_dim: int = 64
    output_dim: int = 1
    lr: float = 0.0003
    prior_mu: int = 0
    prior_sigma: int = 1
    lambda_reg: float = 0.01
    classification_cutoff: float = 0.5


# ----- Maximum Likelihood Regularised EM Parameters ----- >>>>>>>>


@dataclass
class MLREMOptions:
    """
    This class contains the parameters as options
    for the Maximum Likelihood Regularised EM model.
    """

    alpha: float = 0.0  # Regularisation parameter
    beta: float = 1.0  # Scaling parameter
    weight_threshold: float = 1e-3  # Threshold for weight pruning
    max_iterations: int = 300  # Maximum iterations for EM
    tolerance: float = 0.01  # Convergence tolerance



import json
from pathlib import Path

import pandas as pd
import streamlit as st
from st_aggrid import AgGrid, GridOptionsBuilder


def group_plots_by_model_and_type(plots):
    """Group plots by model name and type.

    Args:
        plots (list): List of Path objects representing plot files.

    Returns:
        dict: Dictionary with model names as keys and plot groups (Train/Test/Coefficients/Other) as values.
    """
    plot_groups = {}
    for p in plots:
        if not p.name.endswith(".png"):
            continue

        # Extract model name and plot type from filename
        filename = p.stem.lower()  # Convert to lowercase for case-insensitive matching
        parts = filename.split("-")

        if len(parts) >= 2:
            model_name = parts[0]
            if model_name not in plot_groups:
                plot_groups[model_name] = {
                    "Train": [],
                    "Test": [],
                    "Coefficients": [],
                    "Other": [],
                }

            custom_plot = "custom" in filename

            if custom_plot:
                # If it's a custom plot, add it to the Other category
                plot_groups[model_name]["Other"].append(p)
                continue

            # Check for plot type in the full filename
            if "train" in filename:
                plot_groups[model_name]["Train"].append(p)
            elif "test" in filename:
                plot_groups[model_name]["Test"].append(p)
            elif any(coef in filename for coef in ["coefficients", "coef"]):
                plot_groups[model_name]["Coefficients"].append(p)
            else:
                plot_groups[model_name]["Other"].append(p)

    return plot_groups


def _display_roc_curves(group: dict, cols):
    """Display ROC curves in two columns.

    Args:
        group (dict): Dictionary containing Train and Test plots
        cols: Streamlit columns
    """
    st.markdown("#### ROC Curves")
    for train_plot in [p for p in group["Train"] if "ROC" in p.stem]:
        cols[0].image(str(train_plot))
    for test_plot in [p for p in group["Test"] if "ROC" in p.stem]:
        cols[1].image(str(test_plot))


def _display_confusion_matrices(group: dict, cols):
    """Display confusion matrices in two columns.

    Args:
        group (dict): Dictionary containing Train and Test plots
        cols: Streamlit columns
    """
    st.markdown("#### Confusion Matrices")
    for train_plot in [p for p in group["Train"] if "Confusion" in p.stem]:
        cols[0].image(str(train_plot))
    for test_plot in [p for p in group["Test"] if "Confusion" in p.stem]:
        cols[1].image(str(test_plot))


def _display_regression_plots(group: dict, cols):
    """Display regression plots in two columns.

    Args:
        group (dict): Dictionary containing Train and Test plots
        cols: Streamlit columns
    """
    for train_plot in group["Train"]:
        cols[0].image(str(train_plot))
    for test_plot in group["Test"]:
        cols[1].image(str(test_plot))


def _display_coefficient_plots(group: dict):
    """Display coefficient plots.

    Args:
        group (dict): Dictionary containing Coefficients plots
    """
    st.markdown("#### Model Coefficients")
    st.image(str(group["Coefficients"][0]))


@st.experimental_fragment
def plot_box(plot_dir: Path, box_title: str):
    """Display the plots in the given directory in the UI.

    Args:
        plot_dir (Path): The directory containing the plots.
        box_title (str): The title of the plot box.
    """
    if plot_dir.exists():
        plots = sorted(plot_dir.iterdir(), key=lambda x: x.stat().st_ctime)
        with st.expander(box_title, expanded=len(plots) > 0):
            # Group plots by model and type
            plot_groups = group_plots_by_model_and_type(plots)

            # Display plots by group
            for model_name, group in plot_groups.items():
                # Capitalise first letter of each word in model name
                formatted_model_name = " ".join(
                    word.capitalize() for word in model_name.split()
                )
                st.markdown(f"### {formatted_model_name}")

                # Determine if this is a classification model by checking for ROC curves
                is_classification = any(
                    "ROC" in p.stem for p in group["Train"] + group["Test"]
                )

                cols = st.columns(2)
                if is_classification:
                    # Show ROC curves and confusion matrices if present
                    if any("ROC" in p.stem for p in group["Train"] + group["Test"]):
                        _display_roc_curves(group, cols)
                    if any(
                        "Confusion" in p.stem for p in group["Train"] + group["Test"]
                    ):
                        _display_confusion_matrices(group, cols)
                else:
                    # Show regression plots if present
                    if group["Train"] or group["Test"]:
                        _display_regression_plots(group, cols)

                # Show coefficient plots if present
                if group["Coefficients"]:
                    _display_coefficient_plots(group)

                # Show other plots
                for other_plot in group["Other"]:
                    plot_name = other_plot.name.split("-")[-1].replace(".png", "")
                    st.markdown(f"#### {plot_name.capitalize()}")
                    st.image(str(other_plot))

                st.markdown("---")


@st.experimental_fragment
def plot_box_v2(plot_paths: list[Path], box_title: str):
    """Display the plots given in the list of plot files.

    Args:
        plot_paths (list[Path]): The paths to the plots to display.
        box_title (str): The title of the box.
    """
    with st.expander(box_title, expanded=len(plot_paths) > 0):
        for plot in sorted(plot_paths, key=lambda x: x.stat().st_ctime):
            st.image(str(plot))


@st.experimental_fragment
def display_metrics_table(metrics_path: Path):
    """
    Display a metrics summary table in a Streamlit app.

    Args:
        metrics_path (Path): The path to the metrics file.
    """
    # Check if metrics file exists
    if not metrics_path.exists():
        st.info("No metrics available yet. Train some models first.")
        return

    # Load the metrics from the file
    with open(metrics_path, "r") as f:
        metrics_dict = json.load(f)

    # Prepare data for the table
    rows = []
    for algorithm, results in metrics_dict.items():
        for dataset, metrics in results.items():
            for metric, values in metrics.items():
                row = {
                    "Model": algorithm,
                    "Set": dataset.capitalize(),
                    "Metric": metric,
                    "Mean ± Std": f"{values['mean']:.3f} ± {values['std']:.3f}",
                }
                rows.append(row)

    # Create a DataFrame
    df = pd.DataFrame(rows)

    # Pivot the DataFrame for a cleaner table
    table = df.pivot(
        index=["Model", "Set"], columns="Metric", values="Mean ± Std"
    ).reset_index()
    table = table.set_index("Model")
    table.sort_values(["Model", "Set"], ascending=[True, True], inplace=True)

    # Display the table in Streamlit
    st.write("### Metrics summary")
    st.write(
        "Metrics are the mean (± standard deviation) of all bootstraps (if using the Holdout"
        " data split) or cross-validation folds (if using K-fold data split or"
        " automatic hyper-parameter search)."
    )
    # TODO: This can be moved to a separate function
    # Build Grid Options
    table = table.reset_index()
    gb = GridOptionsBuilder.from_dataframe(table)

    # Apply Global Font Styling to All Columns
    global_style = {
        "fontSize": "14px",
        "fontFamily": "Arial, sans-serif",
        "color": "black",
        "textAlign": "center",
    }

    gb.configure_default_column(
        editable=False, resizable=True, flex=2, cellStyle=global_style, wrapText=False
    )
    gb.configure_first_column_as_index()
    gb.configure_auto_height()
    grid_options = gb.build()

    # Display Full-Width Ag-Grid Table with Styling
    AgGrid(
        table,
        gridOptions=grid_options,
        fit_columns_on_grid_load=True,
        highlight_odd_rows=True,
        key="metrics_table",
        height=150,  # Required to make sure there's not a massive gap below
    )


@st.experimental_fragment
def display_predictions(predictions_df: pd.DataFrame):
    """
    Display the predictions from multiple models in a clean, concatenated format.

    Args:
        predictions_df (pd.DataFrame): DataFrame containing columns ['Model Name', 'Y True', 'Y Prediction', ...]
    """
    st.write("### Predictions")

    # Group by model and reshape each sub-DF
    dfs = []
    for model_name, group_df in predictions_df.groupby("Model Name"):
        model_df = group_df.copy()

        # Rename prediction column to include the model name
        model_df = model_df.rename(columns={"Y Prediction": f"Prediction {model_name}"})

        # Drop model name (already included in prediction column)
        model_df = model_df.drop(columns=["Model Name"])

        dfs.append(model_df.reset_index(drop=True))

    # Concatenate all horizontally and remove duplicated columns
    concatenated = pd.concat(dfs, axis=1)
    concatenated = concatenated.loc[:, ~concatenated.columns.duplicated()]

    # Reorder columns: keep others, then Y True, then predictions
    y_true_col = "Y True"
    prediction_cols = [
        col for col in concatenated.columns if col.startswith("Prediction")
    ]
    other_cols = [
        col for col in concatenated.columns if col not in prediction_cols + [y_true_col]
    ]
    ordered_cols = (
        other_cols
        + ([y_true_col] if y_true_col in concatenated.columns else [])
        + prediction_cols
    )

    st.write(concatenated[ordered_cols])


def fuzzy_feature_impacts(impacts_df: pd.DataFrame):
    """Display the impacts of each feature on each cluster following Fuzzy analysis.

    Args:
        impacts_df (pd.DataFrame): DataFrame containing the impacts of each feature on each cluster.
    """
    with st.expander("Fuzzy Feature Impacts", expanded=True):
        st.dataframe(impacts_df, hide_index=True, use_container_width=True)



from typing import Literal

import streamlit as st

from helix.options.enums import (
    FeatureImportanceStateKeys,
    FuzzyStateKeys,
    MachineLearningStateKeys,
)


@st.experimental_fragment
def log_box(
    box_title: str,
    key: Literal[
        MachineLearningStateKeys.MLLogBox,
        FeatureImportanceStateKeys.FILogBox,
        FuzzyStateKeys.FuzzyLogBox,
    ],
):
    """Display a text area which shows that logs of the current pipeline run."""
    with st.expander(box_title, expanded=False):
        st.text_area(
            box_title,
            key=key,
            height=200,
            disabled=True,
        )



import dataclasses
import json
from pathlib import Path
from typing import TypeVar

from helix.options.data import DataOptions, DataSplitOptions
from helix.options.execution import ExecutionOptions
from helix.options.fi import FeatureImportanceOptions
from helix.options.fuzzy import FuzzyOptions
from helix.options.ml import MachineLearningOptions
from helix.options.plotting import PlottingOptions
from helix.options.preprocessing import PreprocessingOptions

Options = TypeVar(
    "Options",
    DataOptions,
    ExecutionOptions,
    PlottingOptions,
    MachineLearningOptions,
    FeatureImportanceOptions,
    FuzzyOptions,
    PreprocessingOptions,
)


def load_execution_options(path: Path) -> ExecutionOptions:
    """Load experiment execution options from the given path.
    The path will be to a `json` file containing the options.

    Args:
        path (Path): The path the `json` file containing the options.

    Returns:
        ExecutionOptions: The execution options.
    """
    with open(path, "r") as json_file:
        options_json = json.load(json_file)
    options = ExecutionOptions(**options_json)
    return options


def load_plot_options(path: Path) -> PlottingOptions:
    """Load plotting options from the given path.
    The path will be to a `json` file containing the plot options.

    Args:
        path (Path): The path the `json` file containing the options.

    Returns:
        PlottingOptions: The plotting options.
    """
    with open(path, "r") as json_file:
        options_json = json.load(json_file)

    # Migrate old seaborn style names to new ones
    style_migrations = {
        "seaborn": "seaborn-v0_8",
        "seaborn-darkgrid": "seaborn-v0_8-dark",
        "seaborn-whitegrid": "seaborn-v0_8-whitegrid",
        "seaborn-dark": "seaborn-v0_8-dark",
        "seaborn-deep": "seaborn-v0_8",
    }

    if "plot_colour_scheme" in options_json:
        old_style = options_json["plot_colour_scheme"]
        if old_style in style_migrations:
            options_json["plot_colour_scheme"] = style_migrations[old_style]
            # Save the migrated style back to the file
            with open(path, "w") as json_file:
                json.dump(options_json, json_file, indent=4)

    options = PlottingOptions(**options_json)
    return options


def save_options(path: Path, options: Options):
    """Save options to a `json` file at the specified path.

    Args:
        path (Path): The path to the `json` file.
        options (Options): The options to save.
    """
    options_json = dataclasses.asdict(options)
    with open(path, "w") as json_file:
        json.dump(options_json, json_file, indent=4)


def load_fi_options(path: Path) -> FeatureImportanceOptions | None:
    """Load feature importance options.

    Args:
        path (Path): The path to the feature importance options file.

    Returns:
        FeatureImportanceOptions | None: The feature importance options.
    """

    try:
        with open(path, "r") as file:
            fi_json_options = json.load(file)
            fi_options = FeatureImportanceOptions(**fi_json_options)
    except FileNotFoundError:
        fi_options = None
    except TypeError:
        fi_options = None

    return fi_options


def load_fuzzy_options(path: Path) -> FuzzyOptions | None:
    """Load fuzzy options.

    Args:
        path (Path): The path to the fuzzy options file.

    Returns:
        FuzzyOptions | None: The fuzzy options.
    """

    try:
        with open(path, "r") as file:
            fuzzy_json_options = json.load(file)
            fuzzy_options = FuzzyOptions(**fuzzy_json_options)
    except FileNotFoundError:
        fuzzy_options = None
    except TypeError:
        fuzzy_options = None

    return fuzzy_options


def load_data_preprocessing_options(path: Path) -> PreprocessingOptions:
    """Load data preprocessing options from the given path.
    The path will be to a `json` file containing the options.

    Args:
        path (Path): The path the `json` file containing the options.

    Returns:
        PreprocessingOptions: The data preprocessing options.
    """

    try:
        with open(path, "r") as json_file:
            options_json = json.load(json_file)
        preprocessing_options = PreprocessingOptions(**options_json)
    except FileNotFoundError:
        preprocessing_options = None
    except TypeError:
        preprocessing_options = None
    return preprocessing_options


def load_data_options(path: Path) -> DataOptions:
    """Load the data options from the JSON file given in `path`.

    Args:
        path (Path): The path to the JSON file containing the data options.

    Returns:
        DataOptions: The data options.
    """
    with open(path, "r") as json_file:
        options_json: dict = json.load(json_file)
    if split_opts := options_json.get("data_split"):
        options_json["data_split"] = DataSplitOptions(**split_opts)
    options = DataOptions(**options_json)
    return options


def load_ml_options(path: Path) -> MachineLearningOptions | None:
    """Load machine learning options from the given path.
    The path will be to a `json` file containing the options.

    Args:
        path (Path): The path the `json` file containing the options.

    Returns:
        MachineLearningOptions: The machine learning options.
    """
    try:
        with open(path, "r") as file:
            ml_json_options = json.load(file)
            ml_options = MachineLearningOptions(**ml_json_options)
    except FileNotFoundError:
        ml_options = None
    except TypeError:
        ml_options = None

    return ml_options



from itertools import chain

import numpy as np
from scipy.special import xlogy
from sklearn.neural_network import MLPClassifier, MLPRegressor
from sklearn.utils.extmath import safe_sparse_dot


class BRNNMixin:
    """Mixin class containing shared methods for BRNNRegressor and BRNNClassifier."""

    def _compute_reg_term(self):
        """Compute L2 regularization term."""
        total_params = sum(w.size for w in self.coefs_)
        return sum(np.sum(w**2) for w in self.coefs_) / total_params

    def _activation_derivative_func(self):
        """Return the derivative function for the activation."""
        from sklearn.neural_network._base import DERIVATIVES

        return DERIVATIVES[self.activation]

    def _compute_loss_grad(
        self, layer, n_samples, activations, deltas, coef_grads, intercept_grads
    ):
        """Compute gradients for coefficients and intercepts with L2 penalty."""
        coef_grads[layer] = safe_sparse_dot(activations[layer].T, deltas[layer])
        total_params = sum(w.size for w in self.coefs_)
        coef_grads[layer] += self.beta_coef * self.coefs_[layer] / total_params
        coef_grads[layer] /= n_samples
        intercept_grads[layer] = np.mean(deltas[layer], 0)

    def _backprop(self, X, y, activations, deltas, coef_grads, intercept_grads):
        """Backpropagation with adaptive regularization."""
        n_samples = X.shape[0]

        # Forward pass
        activations = self._forward_pass(activations)

        # Compute loss components
        data_loss = self._compute_data_loss(y, activations)
        reg_term = self._compute_reg_term()
        loss = self.alpha_loss * data_loss + self.beta_coef * reg_term

        # Compute output layer deltas
        self._compute_deltas(y, activations, deltas, n_samples)

        # Last layer gradients
        self._compute_loss_grad(
            self.n_layers_ - 2,
            n_samples,
            activations,
            deltas,
            coef_grads,
            intercept_grads,
        )

        # Backpropagate through hidden layers
        inplace_derivative = self._activation_derivative_func()
        for i in range(self.n_layers_ - 2, 0, -1):
            deltas[i - 1] = safe_sparse_dot(deltas[i], self.coefs_[i].T)
            inplace_derivative(activations[i], deltas[i - 1])
            self._compute_loss_grad(
                i - 1, n_samples, activations, deltas, coef_grads, intercept_grads
            )

        # Update alpha and beta adaptively
        Np = sum(p.size for p in chain(self.coefs_, self.intercepts_))
        g_diag = np.array([np.sum(p**2) for p in chain(self.coefs_, self.intercepts_)])
        trace_G_inv = np.sum(1.0 / (self.alpha_loss + g_diag + 1e-8))
        gamma = Np - self.alpha_loss * trace_G_inv
        self.alpha_loss = gamma / (2 * reg_term)
        self.beta_coef = (n_samples - gamma) / (2 * data_loss)

        return loss, coef_grads, intercept_grads

    def _compute_data_loss(self, y, activations):
        """Compute data-specific loss (to be implemented by subclasses)."""
        raise NotImplementedError

    def _compute_deltas(self, y, activations, deltas, n_samples):
        """Compute output layer deltas (to be implemented by subclasses)."""
        raise NotImplementedError


def log_loss(y_true, y_prob, sample_weight=None):
    """Log loss function."""
    eps = np.finfo(y_prob.dtype).eps
    y_prob = np.clip(y_prob, eps, 1 - eps)
    if y_prob.shape[1] == 1:
        y_prob = np.append(1 - y_prob, y_prob, axis=1)
    if y_true.shape[1] == 1:
        y_true = np.append(1 - y_true, y_true, axis=1)
    return -np.average(xlogy(y_true, y_prob), weights=sample_weight, axis=0).sum()


class BRNNRegressor(BRNNMixin, MLPRegressor):
    """BRNN Regressor with RMSE + L2 loss."""

    def __init__(
        self,
        alpha_loss=0.01,
        beta_coef=100,
        hidden_layer_sizes=(100,),
        activation="relu",
        batch_size="auto",
        learning_rate_init=0.001,
        max_iter=200,
        random_state=None,
    ):
        super().__init__(
            hidden_layer_sizes=hidden_layer_sizes,
            activation=activation,
            batch_size=batch_size,
            learning_rate_init=learning_rate_init,
            max_iter=max_iter,
            random_state=random_state,
        )
        self.alpha_loss = alpha_loss
        self.beta_coef = beta_coef
        self.model_name = "BRNNRegressor"

    def _compute_data_loss(self, y, activations):
        errors = activations[-1] - y
        mse = np.mean(errors**2)
        return np.sqrt(mse + 1e-8)

    def _compute_deltas(self, y, activations, deltas, n_samples):
        errors = activations[-1] - y
        rmse = self._compute_data_loss(y, activations)
        deltas[-1] = errors / (n_samples * rmse)


class BRNNClassifier(BRNNMixin, MLPClassifier):
    """BRNN Classifier with log loss + L2 loss."""

    def __init__(
        self,
        alpha_loss=0.01,
        beta_coef=100,
        hidden_layer_sizes=(100,),
        activation="relu",
        batch_size="auto",
        learning_rate_init=0.001,
        max_iter=200,
        random_state=None,
        class_weight=None,
    ):
        super().__init__(
            hidden_layer_sizes=hidden_layer_sizes,
            activation=activation,
            batch_size=batch_size,
            learning_rate_init=learning_rate_init,
            max_iter=max_iter,
            random_state=random_state,
        )
        self.alpha_loss = alpha_loss
        self.beta_coef = beta_coef
        self.model_name = "BRNNClassifier"
        self.class_weight = class_weight

    def _compute_data_loss(self, y, activations):
        y_pred = activations[-1]
        return log_loss(y, y_pred)

    def _compute_deltas(self, y, activations, deltas, n_samples):
        y_pred = activations[-1]
        deltas[-1] = (y_pred - y) / n_samples



from dataclasses import dataclass


@dataclass
class FuzzyOptions:
    cluster_names: list
    fuzzy_feature_selection: bool = True
    number_fuzzy_features: int = 10
    granular_features: bool = True
    number_clusters: int = 5
    number_rules: int = 5
    save_fuzzy_set_plots: bool = True
    fuzzy_log_dir: str = "fuzzy"



import os
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Tuple

import numpy as np
import pandas as pd
import streamlit as st
from sklearn.model_selection import StratifiedKFold, train_test_split
from sklearn.preprocessing import MinMaxScaler, StandardScaler

from helix.options.data import DataOptions
from helix.options.enums import DataSplitMethods, Normalisations, ProblemTypes
from helix.options.execution import ExecutionOptions
from helix.utils.logging_utils import Logger


class DataBuilder:
    """
    Data builder class
    """

    _normalization_dict = {
        Normalisations.MinMax: MinMaxScaler,
        Normalisations.Standardisation: StandardScaler,
    }

    def __init__(
        self,
        data_opts: DataOptions,
        exec_opts: ExecutionOptions,
        logger: object = None,
    ) -> None:
        self._path = data_opts.data_path
        self._data_split = data_opts.data_split
        self._random_state = exec_opts.random_state
        self._logger = logger
        self._normalization = data_opts.normalisation
        self._numerical_cols = "all"
        self._problem_type = exec_opts.problem_type
        self._feature_cols = data_opts.feature_columns
        self._id_col = data_opts.id_column
        self._target_col = data_opts.target_column

    def _load_data(self) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Load data from a csv file

        Returns
        -------
        Tuple[pd.DataFrame, pd.DataFrame]
            The training data (X) and the targets (y)
        """
        df = read_data(Path(self._path), self._logger)
        X = df.iloc[:, :-1]
        y = df.iloc[:, -1]
        return X, y

    def _generate_data_splits(
        self, X: pd.DataFrame, y: pd.DataFrame
    ) -> Dict[str, List[pd.DataFrame]]:
        """Generate data splits for bootstrapping.

        Args:
            X (pd.DataFrame): The training data.
            y (pd.DataFrame): The prediction targets.

        Raises:
            NotImplementedError: Tried to use an unimplemented data split method.

        Returns:
            Dict[str, List[pd.DataFrame]]: The bootstrapped data.
        """
        X_train_list = []
        X_test_list = []
        y_train_list = []
        y_test_list = []
        ids_list = []

        if (
            self._data_split is not None
            and self._data_split.method.lower() == DataSplitMethods.Holdout
        ):
            for i in range(self._data_split.n_bootstraps):
                self._logger.info(
                    "Using holdout data split "
                    f"with test size {self._data_split.test_size} "
                    f"for bootstrap {i+1}"
                )
                if self._problem_type == ProblemTypes.Regression:
                    stratify = None
                elif self._problem_type == ProblemTypes.Classification:
                    stratify = y
                X_train, X_test, y_train, y_test = train_test_split(
                    X,
                    y,
                    test_size=self._data_split.test_size,
                    random_state=self._random_state + i,
                    stratify=stratify,
                    shuffle=True,
                )

                # Assign the data to the lists depending on if there is an ID column
                X_tr_l, X_te_l, y_tr_l, y_te_l, ids_l = self._assign_splits_to_lists(
                    X_train, X_test, y_train, y_test
                )
                X_train_list.extend(X_tr_l)
                X_test_list.extend(X_te_l)
                y_train_list.extend(y_tr_l)
                y_test_list.extend(y_te_l)
                ids_list.extend(ids_l)
        elif (
            self._data_split is not None
            and self._data_split.method.lower() == DataSplitMethods.KFold
        ):
            folds = self._data_split.k_folds
            kf = StratifiedKFold(
                n_splits=folds, shuffle=True, random_state=self._random_state
            )
            kf.get_n_splits(X)

            if self._problem_type == ProblemTypes.Regression:
                stratify = np.zeros(y.shape[0])
            elif self._problem_type == ProblemTypes.Classification:
                stratify = y

            for i, (train_index, test_index) in enumerate(kf.split(X, stratify)):

                self._logger.info(
                    "Using K-Fold data split "
                    f"with test size {len(test_index)} "
                    f"for fold {i+1}"
                )

                X_train, X_test = X.iloc[train_index], X.iloc[test_index]
                y_train, y_test = y.iloc[train_index], y.iloc[test_index]

                # Assign the data to the lists depending on if there is an ID column
                X_tr_l, X_te_l, y_tr_l, y_te_l, ids_l = self._assign_splits_to_lists(
                    X_train, X_test, y_train, y_test
                )
                X_train_list.extend(X_tr_l)
                X_test_list.extend(X_te_l)
                y_train_list.extend(y_tr_l)
                y_test_list.extend(y_te_l)
                ids_list.extend(ids_l)
        elif (
            self._data_split is not None
            and self._data_split.method.lower() == DataSplitMethods.NoSplit
        ):
            if self._problem_type == ProblemTypes.Regression:
                stratify = None
            else:
                stratify = y
            X_train, X_test, y_train, y_test = train_test_split(
                X,
                y,
                test_size=self._data_split.test_size,
                random_state=self._random_state,
                stratify=stratify,
            )
            # Assign the data to the lists depending on if there is an ID column
            X_tr_l, X_te_l, y_tr_l, y_te_l, ids_l = self._assign_splits_to_lists(
                X_train, X_test, y_train, y_test
            )
            X_train_list.extend(X_tr_l)
            X_test_list.extend(X_te_l)
            y_train_list.extend(y_tr_l)
            y_test_list.extend(y_te_l)
            ids_list.extend(ids_l)
        else:
            raise NotImplementedError(
                f"Data split type {self._data_split.method} is not implemented"
            )

        return {
            "X_train": X_train_list,
            "X_test": X_test_list,
            "y_train": y_train_list,
            "y_test": y_test_list,
            "id_column": ids_list if ids_list else None,
        }

    def _normalise_data(
        self,
        data: pd.DataFrame,
    ) -> pd.DataFrame:
        """
        Normalise data using MinMaxScaler

        Parameters
        ----------
        data : pd.DataFrame
            The data to normalise

        Returns
        -------
        X : pd.DataFrame
            Dataframe of normalised data
        """
        if self._normalization.lower() == Normalisations.NoNormalisation:
            return data

        self._logger.info(f"Normalising data using {self._normalization}...")

        scaler = self._normalization_dict.get(self._normalization.lower())
        if not scaler:
            raise ValueError(
                f"Normalization {self._normalization} is not available. "
                f"Choices are {self._normalization_dict.keys()}"
            )
        scaler = scaler()  # create the scaler object

        if isinstance(self._numerical_cols, str) and self._numerical_cols == "all":
            self._numerical_cols = data.columns
        elif isinstance(self._numerical_cols, pd.Index):
            pass
        else:
            raise TypeError("numerical_cols must be a list of columns or 'all'.")
        data[self._numerical_cols] = scaler.fit_transform(data[self._numerical_cols])
        return data

    def _assign_splits_to_lists(self, X_train, X_test, y_train, y_test):
        X_train_list = []
        X_test_list = []
        y_train_list = []
        y_test_list = []
        ids_list = []

        if self._id_col is not None:
            X_train_list.append(X_train[self._feature_cols])
            X_test_list.append(X_test[self._feature_cols])
            y_train_list.append(y_train)
            y_test_list.append(y_test)

            # Combine train and test and select id column
            X_train_test = pd.concat([X_train, X_test], ignore_index=True)
            ids_list.append(X_train_test[self._id_col])
        else:
            X_train_list.append(X_train)
            X_test_list.append(X_test)
            y_train_list.append(y_train)
            y_test_list.append(y_test)
            # Don't add it the ID list if the user didn't give an ID column

        return X_train_list, X_test_list, y_train_list, y_test_list, ids_list

    def ingest(self):
        X, y = self._load_data()
        data = self._generate_data_splits(X, y)

        return TabularData(
            X_train=data["X_train"],
            X_test=data["X_test"],
            y_train=data["y_train"],
            y_test=data["y_test"],
            id_column=data["id_column"],
        )


@dataclass
class TabularData:
    # X_train as a list of dataframes
    X_train: list[pd.DataFrame]
    X_test: list[pd.DataFrame]
    y_train: list[pd.DataFrame]
    y_test: list[pd.DataFrame]
    id_column: list[pd.Series]


@st.cache_data(show_spinner="Loading data...")
def ingest_data(
    exec_opts: ExecutionOptions, data_opts: DataOptions, _logger: Logger
) -> TabularData:
    """
    Load data from disk if the data is not in the streamlit cache,
    else return the stored value. This behaviour is controlled by the
    decorator on the function signature (`@st.cache_data`).

    Args:
        exec_opts (ExecutionOptions): The execution options.
        data_opts (DataOptions): The data options.
        _logger (Logger): The logger.

    Returns:
        TabularData: The ingested data.
    """
    data = DataBuilder(
        data_opts=data_opts,
        exec_opts=exec_opts,
        logger=_logger,
    ).ingest()
    return data


@st.cache_data(show_spinner="Loading data...")
def read_data(data_path: Path, _logger: Logger) -> pd.DataFrame:
    """Read a data file into memory from a '.csv' or '.xlsx' file.

    Args:
        data_path (Path): The path to the file to be read.
        logger (Logger): The logger.

    Raises:
        ValueError: The data file wasn't a '.csv' or '.xlsx' file.

    Returns:
        pd.DataFrame: The data read from the file.
    """
    if data_path.suffix == ".csv":
        try:
            _logger.info(f"Reading data from {data_path}")
            return pd.read_csv(data_path, header=0)
        except Exception as e:
            _logger.error(f"Failed to read data from {data_path}{os.linesep}{e}")
            raise
    elif data_path.suffix == ".xlsx":
        try:
            _logger.info(f"Reading data from {data_path}")
            return pd.read_excel(data_path, header=0)
        except Exception as e:
            _logger.error(f"Failed to read data from {data_path}{os.linesep}{e}")
            raise
    else:
        raise ValueError("data_path must be to a '.csv' or '.xlsx' file")


def read_uploaded_data(uploaded_file) -> pd.DataFrame:
    """Read a data file into memory from a '.csv' or '.xlsx' file.

    Args:
        uploaded_file: The uploaded file to be read.

    Raises:
        ValueError: The data file wasn't a '.csv' or '.xlsx' file.

    Returns:
        pd.DataFrame: The data read from the file.
    """
    if uploaded_file.name.endswith(".csv"):
        try:
            return pd.read_csv(uploaded_file, header=0)
        except Exception as e:
            raise ValueError(f"Failed to read uploaded file{os.linesep}{e}")
    elif uploaded_file.name.endswith(".xlsx"):
        try:
            return pd.read_excel(uploaded_file, header=0)
        except Exception as e:
            raise ValueError(f"Failed to read uploaded file{os.linesep}{e}")
    else:
        raise ValueError("uploaded_file must be a '.csv' or '.xlsx' file")


def rearrange_data(df: pd.DataFrame, data_opts: DataOptions):
    """Rearranges the data frame so that all feature columns are first and the target column last.

    Args:
        df (pd.DataFrame): The data frame to rearrange.
        data_opts (DataOptions): The data options containing the target and feature columns.

    Returns:
        pd.DataFrame: The rearranged data frame.
    """

    target_col = data_opts.target_column
    feature_cols = data_opts.feature_columns
    id_col = data_opts.id_column

    if id_col is not None:
        cols = [id_col] + feature_cols + [target_col]
    else:
        cols = feature_cols + [target_col]

    df = df[cols]

    return df


def save_data(data_path: Path, data: pd.DataFrame, logger: Logger):
    """Save data to either a '.csv' or '.xlsx' file.

    Args:
        data_path (Path): The path to save the data to.
        data (pd.DataFrame): The data to save.
        logger (Logger): The logger.

    Raises:
        ValueError: The data file wasn't a '.csv' or '.xlsx' file.
    """
    if data_path.suffix == ".csv":
        try:
            logger.info(f"Saveing data to {data_path}")
            data.to_csv(data_path, index=False)
        except Exception as e:
            logger.error(f"Failed to save data to {data_path}{os.linesep}{e}")
            raise
    elif data_path.suffix == ".xlsx":
        try:
            logger.info(f"Saving data to {data_path}")
            data.to_excel(data_path, index=False)
        except Exception as e:
            logger.error(f"Failed to save data to {data_path}{os.linesep}{e}")
            raise
    else:
        raise ValueError("data_path must be to a '.csv' or '.xlsx' file")



import argparse
import os
import random
import shutil
from multiprocessing import Process
from pathlib import Path
from typing import Any

import numpy as np
import pandas as pd

from helix.options.data import DataOptions
from helix.services.data import rearrange_data, save_data
from helix.utils.logging_utils import Logger, close_logger


def set_seed(seed: int) -> None:
    """
    Sets the seed for the experiment

    Parameters
    ----------
    seed: int
        The seed to use for the experiment
    """
    # torch.manual_seed(seed)
    np.random.seed(seed)
    random.seed(seed)
    # torch.cuda.manual_seed(seed)


def log_options(log_directory, opt: argparse.Namespace):
    """Log model or feature importance hyperparameters
    Parameters
     ----------
         log_directory: str
             The directory to save the log file
         opt: argparse.Namespace
             The options object
     Returns:
         None
    """

    log_path = os.path.join(log_directory, "options.txt")

    with open(log_path, "w") as f:
        for arg in vars(opt):
            f.write(f"{arg}: {getattr(opt, arg)}\n")


def create_directory(path: Path):
    """Create a directory at the specified path. If intermediate directories
    don't already exist, create them. If the path already exists, no action
    is taken.

    Args:
        path (Path): The path the directory to create.
    """
    if not os.path.exists(path):
        os.makedirs(path, exist_ok=True)


def save_upload(
    upload_path: str, content: Any, data_options: DataOptions, mode: str = "w"
):
    """Save a file given to the UI to disk.

    Args:
        upload_path (str): The path to save file to.
        content (Any): The output from `st.file_uploader`.
        mode (str): The mode to write the file. e.g. "w", "w+", "wb", etc.
    """
    logger_instance = Logger()
    logger = logger_instance.make_logger()

    base_dir = os.path.dirname(upload_path)
    if not os.path.exists(base_dir):
        os.makedirs(base_dir)

    if content.name.endswith(".csv"):
        df = pd.read_csv(content, header=0)

    elif content.name.endswith(".xlsx"):
        df = pd.read_excel(content, header=0)
    else:
        raise ValueError("Uploaded file must be a '.csv' or a '.xlsx' file.")

    df = rearrange_data(df, data_options)
    save_data(upload_path, df, logger)
    close_logger(logger_instance, logger)


def cancel_pipeline(p: Process):
    """Cancel a running pipeline.

    Args:
        p (Process): the process running the pipeline to cancel.
    """
    if p.is_alive():
        p.terminate()


def delete_directory(path: Path):
    """Delete a directory and all of its contents.

    Returns:
        None

    Args:
        path: The path to the log directory
    """
    if path.is_dir():
        shutil.rmtree(path, ignore_errors=True)  # Remove the directory


def delete_file(path: Path):
    """Delete a file.

    Returns:
        None

    Args:
        path: The path to the log directory
    """
    if path.is_file():
        path.unlink()  # Remove the file



import os
from pathlib import Path

from helix.options.data import DataOptions
from helix.options.execution import ExecutionOptions
from helix.options.file_paths import (
    data_options_path,
    execution_options_path,
    fi_options_dir,
    fi_options_path,
    fi_plot_dir,
    fi_result_dir,
    fuzzy_options_path,
    fuzzy_plot_dir,
    fuzzy_result_dir,
    helix_experiments_base_dir,
    log_dir,
    plot_options_path,
)
from helix.options.plotting import PlottingOptions
from helix.services.configuration import save_options
from helix.utils.utils import create_directory, delete_directory, delete_file


def get_experiments(base_dir: Path | None = None) -> list[str]:
    """Get the list of experiments in the Helix experiment directory.

    If `base_dir` is not specified, the default from `helix_experiments_base_dir`
    is used

    Args:
        base_dir (Path | None, optional): Specify a base directory for experiments.
        Defaults to None.

    Returns:
        list[str]: The list of experiments.
    """
    # Get the base directory of all experiments
    if base_dir is None:
        base_dir = helix_experiments_base_dir()

    if not base_dir.exists():
        # if no experiments directory, return empty list
        return []
    experiments = os.listdir(base_dir)
    # Filter out hidden files and directories
    experiments = filter(lambda x: not x.startswith("."), experiments)
    # Filter out files
    experiments = filter(
        lambda x: os.path.isdir(os.path.join(base_dir, x)), experiments
    )
    return sorted(experiments)


def create_experiment(
    save_dir: Path,
    plotting_options: PlottingOptions,
    execution_options: ExecutionOptions,
    data_options: DataOptions,
):
    """Create an experiment on disk with it's global plotting options,
    execution options and data options saved as `json` files.

    Args:
        save_dir (Path): The path to where the experiment will be created.
        plotting_options (PlottingOptions): The plotting options to save.
        execution_options (ExecutionOptions): The execution options to save.
        data_options (DataOptions): The data options to save.
    """
    create_directory(save_dir)
    plot_file_path = plot_options_path(save_dir)
    save_options(plot_file_path, plotting_options)
    execution_file_path = execution_options_path(save_dir)
    save_options(execution_file_path, execution_options)
    data_file_path = data_options_path(save_dir)
    save_options(data_file_path, data_options)


def find_previous_fi_results(experiment_path: Path) -> bool:
    """Find previous feature importance results.

    Args:
        experiment_path (Path): The path to the experiment.

    Returns:
        bool: whether previous experiments exist or not.
    """

    directories = [
        fi_plot_dir(experiment_path),
        fi_result_dir(experiment_path),
        fi_options_dir(experiment_path),
        fuzzy_plot_dir(experiment_path),
        fuzzy_result_dir(experiment_path),
        fuzzy_options_path(experiment_path),
        fi_options_path(experiment_path),
        log_dir(experiment_path) / "fi",
        log_dir(experiment_path) / "fuzzy",
    ]

    return any([d.exists() for d in directories])


def delete_previous_fi_results(experiment_path: Path):
    """Delete previous feature importance results.

    Args:
        experiment_path (Path): The path to the experiment.
    """

    directories = [
        fi_plot_dir(experiment_path),
        fi_result_dir(experiment_path),
        fi_options_dir(experiment_path),
        fuzzy_plot_dir(experiment_path),
        fuzzy_result_dir(experiment_path),
        fuzzy_options_path(experiment_path),
        fi_options_path(experiment_path),
        log_dir(experiment_path) / "fi",
        log_dir(experiment_path) / "fuzzy",
    ]

    for directory in directories:
        if directory.exists():
            if directory.is_file():
                delete_file(directory)
            elif directory.is_dir():
                delete_directory(directory)



from matplotlib import pyplot as plt
from matplotlib.figure import Figure


def close_figure(figure: Figure):
    """Clear and close a Figure that is no longer in use.

    A good time to use this is after saving a Figure to disk.

    Args:
        figure (Figure): The Figure to close.
    """
    figure.clear()
    plt.close(figure)



from dataclasses import dataclass


@dataclass
class FeatureImportanceOptions:
    global_importance_methods: dict
    feature_importance_ensemble: dict
    local_importance_methods: dict
    save_feature_importance_results: bool = True
    save_feature_importance_options: bool = True
    save_feature_importance_plots: bool = True
    num_features_to_plot: int = 5
    permutation_importance_scoring: str = "neg_mean_absolute_error"
    permutation_importance_repeat: int = 10
    fi_log_dir: str = "fi"



import json
import os
from pathlib import Path

import pandas as pd

from helix.options.data import DataOptions
from helix.options.enums import FeatureImportanceTypes, Metrics, ProblemTypes
from helix.options.execution import ExecutionOptions
from helix.options.fi import FeatureImportanceOptions
from helix.options.file_paths import (
    fi_plot_dir,
    fi_result_dir,
    helix_experiments_base_dir,
    ml_metrics_full_path,
    ml_metrics_mean_std_path,
)
from helix.options.plotting import PlottingOptions
from helix.services.data import TabularData, read_data
from helix.services.feature_importance.ensemble_methods import (
    calculate_ensemble_majorityvote,
    calculate_ensemble_mean,
)
from helix.services.feature_importance.global_methods import (
    calculate_global_shap_values,
    calculate_permutation_importance,
)
from helix.services.feature_importance.local_methods import (
    calculate_lime_values,
    calculate_local_shap_values,
)
from helix.services.metrics import find_mean_model_index
from helix.services.plotting import (
    plot_bar_chart,
    plot_global_shap_importance,
    plot_lime_importance,
    plot_local_shap_importance,
    plot_permutation_importance,
)
from helix.utils.logging_utils import Logger
from helix.utils.plotting import close_figure
from helix.utils.utils import create_directory


class FeatureImportanceEstimator:
    """
    Interpreter class to interpret the model results.

    """

    def __init__(
        self,
        fi_opt: FeatureImportanceOptions,
        exec_opt: ExecutionOptions,
        plot_opt: PlottingOptions,
        data_opt: DataOptions,
        data_path: Path,
        logger: Logger | None = None,
    ) -> None:
        self._fi_opt = fi_opt
        self._logger = logger
        self._exec_opt = exec_opt
        self._plot_opt = plot_opt
        self._data_opt = data_opt
        self._feature_importance_methods = self._fi_opt.global_importance_methods
        self._local_importance_methods = self._fi_opt.local_importance_methods
        self._feature_importance_ensemble = self._fi_opt.feature_importance_ensemble
        self._data_path = data_path

    def interpret(self, models: dict, data: TabularData) -> tuple[dict, dict, dict]:
        """
        Interpret the model results using the selected feature importance methods
        and ensemble methods.
        Parameters:
            models (dict): Dictionary of models.
            data (TabularData): The data to interpret.
        Returns:
            tuple[dict, dict, dict]:
            Global, local and ensemble feature importance votes.
        """
        self._logger.info("-------- Start of feature importance logging--------")
        global_importance_results = self._global_feature_importance(models, data)
        global_importance_df_dict = self._stack_importances(global_importance_results)
        # Compute average global importance across all folds for each model type
        self._calculate_mean_global_importance_of_folds(global_importance_results)

        # Load the total dataset for the local importance
        total_df = read_data(self._data_path, self._logger)
        if self._data_opt.id_column is not None:
            total_df.drop(self._data_opt.id_column, axis=1, inplace=True)
        local_importance_results = self._local_feature_importance(models, total_df)
        ensemble_results = self._ensemble_feature_importance(global_importance_df_dict)
        self._logger.info("-------- End of feature importance logging--------")

        return global_importance_df_dict, local_importance_results, ensemble_results

    def _global_feature_importance(self, models: dict, data: TabularData):
        """
        Calculate global feature importance for a given model and dataset.
        Parameters:
            models (dict): Dictionary of models.
            data (TabularData): The data to interpret.
        Returns:
            dict: Dictionary of feature importance results.
        """
        feature_importance_results = {}
        if not any(
            sub_dict["value"] for sub_dict in self._feature_importance_methods.values()
        ):
            self._logger.info("No feature importance methods selected")
            self._logger.info("Skipping global feature importance methods")
            return feature_importance_results

        # Iterate through all data indices
        for idx in range(len(data.X_train)):
            X, y = data.X_train[idx], data.y_train[idx]

            # Iterate through all models
            for model_type, model_list in models.items():
                self._logger.info(
                    f"Global feature importance methods for {model_type} (fold {idx + 1})..."
                )
                if model_type not in feature_importance_results:
                    feature_importance_results[model_type] = {}

                # Iterate through all feature importance methods
                for (
                    feature_importance_type,
                    value,
                ) in self._feature_importance_methods.items():
                    if not value["value"]:
                        continue

                    if (
                        feature_importance_type
                        not in feature_importance_results[model_type]
                    ):
                        feature_importance_results[model_type][
                            feature_importance_type
                        ] = []

                    if (
                        feature_importance_type
                        == FeatureImportanceTypes.PermutationImportance
                    ):
                        # Run Permutation Importance
                        permutation_importance_df = calculate_permutation_importance(
                            model=model_list[idx],
                            X=X,
                            y=y,
                            permutation_importance_scoring=self._fi_opt.permutation_importance_scoring,
                            permutation_importance_repeat=self._fi_opt.permutation_importance_repeat,
                            random_state=self._exec_opt.random_state,
                            logger=self._logger,
                        )
                        results_dir = fi_result_dir(
                            helix_experiments_base_dir()
                            / self._exec_opt.experiment_name
                        )
                        create_directory(results_dir)
                        permutation_importance_df.to_csv(
                            results_dir
                            / f"global-{feature_importance_type}-{model_type}-fold-{idx + 1}.csv"
                        )
                        fig = plot_permutation_importance(
                            permutation_importance_df,
                            self._plot_opt,
                            self._fi_opt.num_features_to_plot,
                            f"{feature_importance_type} - {model_type} (fold {idx + 1})",
                        )
                        plot_dir = fi_plot_dir(
                            helix_experiments_base_dir()
                            / self._exec_opt.experiment_name
                        )
                        create_directory(plot_dir)
                        fig.savefig(
                            plot_dir
                            / f"{feature_importance_type}-{value['type']}-{model_type}-fold-{idx + 1}-bar.png"
                        )
                        close_figure(fig)
                        feature_importance_results[model_type][
                            feature_importance_type
                        ].append(permutation_importance_df)

                    elif feature_importance_type == FeatureImportanceTypes.SHAP:
                        # Run SHAP
                        shap_df, _ = calculate_global_shap_values(
                            model=model_list[idx],
                            X=X,
                            logger=self._logger,
                        )
                        results_dir = fi_result_dir(
                            helix_experiments_base_dir()
                            / self._exec_opt.experiment_name
                        )
                        create_directory(results_dir)
                        shap_df.to_csv(
                            results_dir
                            / f"global-{feature_importance_type}-{model_type}-fold-{idx + 1}.csv"
                        )
                        fig = plot_global_shap_importance(
                            shap_values=shap_df,
                            plot_opts=self._plot_opt,
                            num_features_to_plot=self._fi_opt.num_features_to_plot,
                            title=f"{feature_importance_type} - {value['type']} - {model_type} (fold {idx + 1})",
                        )
                        plot_dir = fi_plot_dir(
                            helix_experiments_base_dir()
                            / self._exec_opt.experiment_name
                        )
                        create_directory(plot_dir)
                        fig.savefig(
                            plot_dir
                            / f"{feature_importance_type}-{value['type']}-{model_type}-fold-{idx + 1}-bar.png"
                        )
                        close_figure(fig)
                        feature_importance_results[model_type][
                            feature_importance_type
                        ].append(shap_df)

        return feature_importance_results

    def _local_feature_importance(self, models, data: pd.DataFrame):
        """
        Calculate local feature importance for a given model and dataset.
        Parameters:
            models (dict): Dictionary of models.
            data (pd.DataFrame): The data to interpret.
            For local interpretation, the entire data is used.
        Returns:
            dict: Dictionary of feature importance results.
        """
        # Get data features
        X = data.iloc[:, :-1]
        # Get the targets
        y = data.iloc[:, -1]

        # Determine which metric to use
        if self._exec_opt.problem_type == ProblemTypes.Regression:
            metric = Metrics.R2.value
        elif self._exec_opt.problem_type == ProblemTypes.Classification:
            metric = Metrics.ROC_AUC.value

        # Load the full ml_metrics
        path_to_metrics = ml_metrics_full_path(
            helix_experiments_base_dir() / self._exec_opt.experiment_name
        )
        # Load the metrics mean and std from the file
        with open(path_to_metrics, "r") as f:
            metrics_full = json.load(f)

        # Load the ml_metrics mean std
        path_to_metrics = ml_metrics_mean_std_path(
            helix_experiments_base_dir() / self._exec_opt.experiment_name
        )
        # Load the metrics mean and std from the file
        with open(path_to_metrics, "r") as f:
            metrics_mean_std = json.load(f)

        feature_importance_results = {}
        if not any(
            sub_dict["value"] for sub_dict in self._local_importance_methods.values()
        ):
            self._logger.info("No local feature importance methods selected")
            self._logger.info("Skipping local feature importance methods")
        else:
            for model_type, model in models.items():
                self._logger.info(
                    f"Local feature importance methods for {model_type}..."
                )
                feature_importance_results[model_type] = {}

                # Get the index for the model closest to the mean performance
                closest_index = find_mean_model_index(
                    metrics_full, metrics_mean_std, metric
                )

                # Run methods with TRUE values in the dictionary of feature importance methods
                for (
                    feature_importance_type,
                    value,
                ) in self._local_importance_methods.items():
                    if value["value"]:
                        # Select the first model in the list - model[0]
                        if feature_importance_type == FeatureImportanceTypes.LIME:
                            # Run Permutation Importance
                            lime_importance_df = calculate_lime_values(
                                model[closest_index],
                                X,
                                self._exec_opt.problem_type,
                                self._logger,
                            )
                            results_dir = fi_result_dir(
                                helix_experiments_base_dir()
                                / self._exec_opt.experiment_name
                            )
                            create_directory(results_dir)
                            lime_importance_df.to_csv(
                                results_dir / f"local-{feature_importance_type}.csv"
                            )
                            fig = plot_lime_importance(
                                df=lime_importance_df,
                                plot_opts=self._plot_opt,
                                num_features_to_plot=self._fi_opt.num_features_to_plot,
                                title=f"{feature_importance_type} - {model_type}",
                            )
                            plot_dir = fi_plot_dir(
                                helix_experiments_base_dir()
                                / self._exec_opt.experiment_name
                            )
                            create_directory(
                                plot_dir
                            )  # will create the directory if it doesn't exist
                            fig.savefig(
                                plot_dir
                                / f"local-{feature_importance_type}-{model_type}-violin.png"
                            )
                            close_figure(fig)

                            # Add class to local feature importance for fuzzy
                            lime_importance_df = pd.concat(
                                [lime_importance_df, y], axis=1
                            )
                            feature_importance_results[model_type][
                                feature_importance_type
                            ] = lime_importance_df

                        if feature_importance_type == FeatureImportanceTypes.SHAP:
                            # Run SHAP
                            shap_df, shap_values = calculate_local_shap_values(
                                model=model[closest_index],
                                X=X,
                                logger=self._logger,
                            )
                            results_dir = fi_result_dir(
                                helix_experiments_base_dir()
                                / self._exec_opt.experiment_name
                            )
                            create_directory(results_dir)
                            shap_df.to_csv(
                                results_dir / f"local-{feature_importance_type}.csv"
                            )
                            fig = plot_local_shap_importance(
                                shap_values=shap_values,
                                plot_opts=self._plot_opt,
                                num_features_to_plot=self._fi_opt.num_features_to_plot,
                                title=f"{feature_importance_type} - {value['type']} - {model_type}",
                            )
                            plot_dir = fi_plot_dir(
                                helix_experiments_base_dir()
                                / self._exec_opt.experiment_name
                            )
                            create_directory(
                                plot_dir
                            )  # will create the directory if it doesn't exist
                            fig.savefig(
                                plot_dir
                                / f"local-{feature_importance_type}-{value['type']}-{model_type}-beeswarm.png"
                            )
                            close_figure(fig)

                            # Add class to local feature importance for fuzzy
                            shap_df = pd.concat([shap_df, y], axis=1)
                            feature_importance_results[model_type][
                                feature_importance_type
                            ] = shap_df

        return feature_importance_results

    def _ensemble_feature_importance(self, feature_importance_results):
        """
        Calculate ensemble feature importance methods.
        Parameters:
            feature_importance_results (dict): Dictionary of feature importance results.
        Returns:
            dict: Dictionary of ensemble feature importance results.
        """
        ensemble_results = {}

        if not any(self._feature_importance_ensemble.values()):
            self._logger.info("No ensemble feature importance method selected")
            self._logger.info("Skipping ensemble feature importance analysis")
        else:
            self._logger.info("Ensemble feature importance methods...")
            for ensemble_type, value in self._feature_importance_ensemble.items():
                if value:
                    if ensemble_type == FeatureImportanceTypes.Mean:
                        # Calculate mean of feature importance results
                        mean_results, mean_results_std = calculate_ensemble_mean(
                            feature_importance_results, self._logger
                        )
                        results_dir = fi_result_dir(
                            helix_experiments_base_dir()
                            / self._exec_opt.experiment_name
                        )
                        create_directory(results_dir)
                        mean_results.to_csv(
                            results_dir / f"ensemble-{ensemble_type}.csv"
                        )
                        fig = plot_bar_chart(
                            df=mean_results,
                            sort_key="Mean Importance",
                            plot_opts=self._plot_opt,
                            title=f"Ensemble {ensemble_type}",
                            x_label="Feature",
                            y_label="Importance",
                            n_features=self._fi_opt.num_features_to_plot,
                            error_bars=mean_results_std,
                        )
                        plot_dir = fi_plot_dir(
                            helix_experiments_base_dir()
                            / self._exec_opt.experiment_name
                        )
                        create_directory(
                            plot_dir
                        )  # will create the directory if it doesn't exist
                        fig.savefig(plot_dir / f"ensemble-{ensemble_type}.png")
                        close_figure(fig)
                        ensemble_results[ensemble_type] = mean_results

                    if ensemble_type == FeatureImportanceTypes.MajorityVote:
                        # Calculate majority vote of feature importance results
                        majority_vote_results, majority_vote_results_std = (
                            calculate_ensemble_majorityvote(
                                feature_importance_results, self._logger
                            )
                        )
                        results_dir = fi_result_dir(
                            helix_experiments_base_dir()
                            / self._exec_opt.experiment_name
                        )
                        create_directory(results_dir)
                        majority_vote_results.to_csv(
                            results_dir / f"ensemble-{ensemble_type}.csv"
                        )
                        fig = plot_bar_chart(
                            df=majority_vote_results,
                            sort_key="Majority Vote Importance",
                            plot_opts=self._plot_opt,
                            title=f"Ensemble {ensemble_type}",
                            x_label="Feature",
                            y_label="Importance",
                            n_features=self._fi_opt.num_features_to_plot,
                            error_bars=majority_vote_results_std,
                        )
                        plot_dir = fi_plot_dir(
                            helix_experiments_base_dir()
                            / self._exec_opt.experiment_name
                        )
                        create_directory(
                            plot_dir
                        )  # will create the directory if it doesn't exist
                        fig.savefig(plot_dir / f"ensemble-{ensemble_type}.png")
                        close_figure(fig)
                        ensemble_results[ensemble_type] = majority_vote_results

            self._logger.info(
                f"Ensemble feature importance results: {os.linesep}{ensemble_results}"
            )

        return ensemble_results

    def _stack_importances(
        self, importances: dict[str, dict[str, list[pd.DataFrame]]]
    ) -> dict[str, pd.DataFrame]:
        """Stack and normalise feature importance results from different methods.

        This function processes feature importance results through these steps:
            - For each model:
           - For each importance type (e.g., SHAP, Permutation):
              - Concatenate all fold results vertically into a single DataFrame
              - Min-max normalise the importance scores to [0,1] range
           - Concatenate all normalised importance types horizontally

        Args:
            importances: Nested dictionary structure:
                - First level: Model name -> Dictionary of importance types
                - Second level: Importance type -> List of DataFrames (one per fold)
                Each DataFrame contains feature importance scores

        Returns:
            Dictionary mapping model names to their stacked importances.
            Each DataFrame has features as rows and importance methods as columns,
            with normalised importance scores as values.
        """
        stack_importances = {}
        for model_name, importance_dict in importances.items():
            importance_type_df_list = []
            for importances_dfs in importance_dict.values():
                importance_df = pd.concat(importances_dfs, axis=0)
                importance_df = (importance_df - importance_df.min()) / (
                    importance_df.max() - importance_df.min()
                )
                importance_type_df_list.append(importance_df)

            stack_importances[model_name] = pd.concat(importance_type_df_list, axis=1)

        return stack_importances

    def _calculate_mean_global_importance_of_folds(
        self, global_importances_dict: dict[str, dict[str, list[pd.DataFrame]]]
    ):
        """Calculate the mean global importance for all folds through which a model was trained.
        The all-folds mean for each model and importance type is saved along with a plot.

        Args:
            global_importances_dict (dict[str, dict[str, list[pd.DataFrame]]]):
                The global importance results containing the importance calculations for
                each model type, importance type and folds.
        """
        for model_name, gfi_dict in global_importances_dict.items():
            for fi_type, importance_dfs in gfi_dict.items():
                fold_mean_df = pd.concat(importance_dfs).groupby(level=0).mean()
                fold_std_df = pd.concat(importance_dfs).groupby(level=0).std()
                results_dir = fi_result_dir(
                    helix_experiments_base_dir() / self._exec_opt.experiment_name
                )
                create_directory(results_dir)
                fold_mean_df.to_csv(
                    results_dir / f"global-{fi_type}-{model_name}-all-folds-mean.csv"
                )
                fig = plot_bar_chart(
                    df=fold_mean_df,
                    sort_key=fold_mean_df.columns[
                        0
                    ],  # there's one column which is the FI type
                    plot_opts=self._plot_opt,
                    title=f"{fi_type} - {model_name} - all folds mean",
                    x_label="Feature",
                    y_label="Importance",
                    n_features=self._fi_opt.num_features_to_plot,
                    error_bars=fold_std_df,
                )
                plot_dir = fi_plot_dir(
                    helix_experiments_base_dir() / self._exec_opt.experiment_name
                )
                create_directory(
                    plot_dir
                )  # will create the directory if it doesn't exist
                fig.savefig(plot_dir / f"{fi_type}-{model_name}-all-folds-mean.png")



from multiprocessing import cpu_count
from time import time
from typing import Any, Tuple

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from sklearn.metrics import make_scorer
from sklearn.model_selection import GridSearchCV, StratifiedKFold

from helix.options.choices.metrics import CLASSIFICATION_METRICS, REGRESSION_METRICS
from helix.options.data import DataSplitOptions
from helix.options.enums import DataSplitMethods, Metrics, ProblemTypes
from helix.services.data import TabularData
from helix.services.metrics import get_metrics
from helix.services.ml_models import get_model, get_model_type
from helix.utils.logging_utils import Logger

_MAX_CPUS = max(cpu_count() - 1, 1)


class Learner:
    """
    Learner class encapsulates the logic for initialising,
    training, and evaluating machine learning models.

    Args:
        - model_types (dict): Dictionary containing model types
        and their parameters.
        - problem_type (ProblemTypes): Type of problem (
        classification or regression).
        - data_split (DataSplitOptions): Object containing data split type and parameters
        - logger (Logger): Logger object to log messages
    """

    def __init__(
        self,
        model_types: dict,
        problem_type: ProblemTypes,
        data_split: DataSplitOptions,
        logger: Logger | None = None,
        n_cpus: int = _MAX_CPUS,
    ) -> None:
        self._logger = logger
        self._model_types = model_types
        self._problem_type = problem_type
        self._data_split = data_split
        self._metrics = get_metrics(self._problem_type, logger=self._logger)
        self._n_cpus = n_cpus

        self._logger.info(f"Using {self._n_cpus} of {cpu_count()} CPUs for training")

    def _process_data_for_bootstrap(self, data, i):
        """
        Extracts and converts the datasets for the given samples
        if it is not a numpy array.

        Args:
            - data: Structured object containing `X_train`,
            `X_test`, `y_train`, `y_test`.
            - i (int): Index of the sample.

        Returns:
            - Tuple: Processed `X_train`, `X_test`, `y_train`,
            `y_test` for the given sample.
        """
        try:
            X_train = (
                data.X_train[i].to_numpy()
                if isinstance(data.X_train[i], pd.DataFrame)
                else data.X_train[i]
            )
            X_test = (
                data.X_test[i].to_numpy()
                if isinstance(data.X_test[i], pd.DataFrame)
                else data.X_test[i]
            )
            y_train = (
                data.y_train[i].to_numpy()
                if isinstance(data.y_train[i], pd.DataFrame)
                else data.y_train[i]
            )
            y_test = (
                data.y_test[i].to_numpy()
                if isinstance(data.y_test[i], pd.DataFrame)
                else data.y_test[i]
            )
        except Exception as e:
            raise ValueError(
                f"Error processing bootstrap data during train test split  {e}"
            )

        return X_train, X_test, y_train, y_test

    def fit(self, data: Tuple):
        """
        Fits machine learning models to the given data
        and applies holdout strategy with bootstrap sampling.

        Args:
            - data (Tuple): Tuple containing training and testing data
            for each bootstrap sample.

        Returns:
            - res (Dict): Dictionary containing model predictions for
            each bootstrap sample.
            - metrics_full (Dict): Dictionary containing metric values for
            each bootstrap sample.
            - metrics_mean_std (Dict): Dictionary containing average and
            standard deviation of metric values across bootstrap samples.
            - trained_models (Dict): Dictionary containing
            trained models for each model type.
        """
        if self._data_split.method.lower() == DataSplitMethods.Holdout:
            res, metrics_full, metrics_mean_std, trained_models = self._fit_holdout(
                data
            )
            return res, metrics_full, metrics_mean_std, trained_models

        elif self._data_split.method.lower() == DataSplitMethods.KFold:
            res, metrics_full, metrics_mean_std, trained_models = self._fit_kfold(data)
            return res, metrics_full, metrics_mean_std, trained_models

    def _fit_holdout(self, data: Tuple) -> None:
        """
        Trains models using a holdout strategy with bootstrap sampling.

        Args:
            - data (Tuple): Tuple containing training and testing data
            for each bootstrap sample.

        Returns:
            - res (Dict): Dictionary containing model predictions for
            each bootstrap sample.
            - metrics_full (Dict): Dictionary containing metric values for
            each bootstrap sample.
            - metrics_mean_std (Dict): Dictionary containing average and
            standard deviation of metric values across bootstrap samples.
            - trained_models (Dict): Dictionary containing
            trained models for each model type.
        """
        self._logger.info("Fitting holdout with bootstrapped datasets...")
        res = {}
        metrics_full = {}
        trained_models = {model_name: [] for model_name in self._model_types.keys()}

        def _fit_single_model_holdout(
            model_name: str, params: dict, X_train, X_test, y_train, y_test
        ):
            model_res = {}
            model_type = get_model_type(model_name, self._problem_type)
            model = get_model(model_type, params["params"])
            self._logger.info(f"Fitting {model_name}...")

            # Standard model handling
            model.fit(X_train, y_train)
            y_pred_train = model.predict(X_train)
            model_res["y_pred_train"] = y_pred_train
            y_pred_test = model.predict(X_test)
            model_res["y_pred_test"] = y_pred_test

            if self._problem_type == ProblemTypes.Classification:
                y_pred_probs_train = model.predict_proba(X_train)
                model_res["y_pred_train_proba"] = y_pred_probs_train
                y_pred_probs_test = model.predict_proba(X_test)
                model_res["y_pred_test_proba"] = y_pred_probs_test
            else:
                y_pred_probs_train = None
                y_pred_probs_test = None

            model_metrics = _evaluate(
                model_name,
                self._metrics,
                y_train,
                y_pred_train,
                y_pred_probs_train,
                y_test,
                y_pred_test,
                y_pred_probs_test,
                self._logger,
                self._problem_type,
            )

            return model_name, model_res, model_metrics, model

        training_start = time()
        for i in range(self._data_split.n_bootstraps):
            self._logger.info(f"\n\n PROCESSING BOOSTRAP NUMBER {i+1}...")
            X_train, X_test, y_train, y_test = self._process_data_for_bootstrap(data, i)

            res[i] = {}

            results = Parallel(n_jobs=self._n_cpus, prefer="processes")(
                delayed(_fit_single_model_holdout)(
                    model_name, params, X_train, X_test, y_train, y_test
                )
                for model_name, params in self._model_types.items()
            )

            for model_name, model_res, model_metrics, model in results:
                res[i][model_name] = model_res
                if model_name not in metrics_full:
                    metrics_full[model_name] = []
                metrics_full[model_name].append(model_metrics)
                trained_models[model_name].append(model)

        training_end = time()
        elapsed = training_end - training_start
        hours = int(elapsed) // 3600
        minutes = (int(elapsed) % 3600) // 60
        seconds = int(elapsed) % 60
        # Create format hh:mm:ss
        time_str = f"{hours:02d}:{minutes:02d}:{seconds:02d}"
        self._logger.info(f"Training completed in {time_str}")

        metrics_mean_std = _compute_metrics_mean_std(metrics_full)
        return res, metrics_full, metrics_mean_std, trained_models

    def _fit_kfold(self, data: Tuple) -> None:
        self._logger.info("Fitting cross validation datasets...")
        res = {}
        metrics_full = {}
        trained_models = {model_name: [] for model_name in self._model_types.keys()}

        def _fit_single_model_kfold(
            model_name: str, params: dict, X_train, X_test, y_train, y_test
        ):
            model_res = {}
            self._logger.info(f"Fitting {model_name}...")
            model_type = get_model_type(model_name, self._problem_type)
            model = get_model(model_type, params["params"])

            # Standard model handling
            model.fit(X_train, y_train)
            y_pred_train = model.predict(X_train)
            model_res["y_pred_train"] = y_pred_train
            y_pred_test = model.predict(X_test)
            model_res["y_pred_test"] = y_pred_test

            if self._problem_type == ProblemTypes.Classification:
                y_pred_probs_train = model.predict_proba(X_train)
                model_res["y_pred_train_proba"] = y_pred_probs_train
                y_pred_probs_test = model.predict_proba(X_test)
                model_res["y_pred_test_proba"] = y_pred_probs_test
            else:
                y_pred_probs_train = None
                y_pred_probs_test = None

            model_metrics = _evaluate(
                model_name,
                self._metrics,
                y_train,
                y_pred_train,
                y_pred_probs_train,
                y_test,
                y_pred_test,
                y_pred_probs_test,
                self._logger,
                self._problem_type,
            )

            return model_name, model_res, model_metrics, model

        training_start = time()
        for i in range(self._data_split.k_folds):
            self._logger.info(f"Processing test fold sample {i+1}...")
            X_train, X_test = data.X_train[i].to_numpy(), data.X_test[i].to_numpy()
            y_train, y_test = data.y_train[i].to_numpy(), data.y_test[i].to_numpy()

            res[i] = {}

            results = Parallel(n_jobs=self._n_cpus, prefer="processes")(
                delayed(_fit_single_model_kfold)(
                    model_name, params, X_train, X_test, y_train, y_test
                )
                for model_name, params in self._model_types.items()
            )

            for model_name, model_res, model_metrics, model in results:
                res[i][model_name] = model_res
                if model_name not in metrics_full:
                    metrics_full[model_name] = []
                metrics_full[model_name].append(model_metrics)
                trained_models[model_name].append(model)

        training_end = time()
        elapsed = training_end - training_start
        hours = int(elapsed) // 3600
        minutes = (int(elapsed) % 3600) // 60
        seconds = int(elapsed) % 60
        # Create format hh:mm:ss
        time_str = f"{hours:02d}:{minutes:02d}:{seconds:02d}"
        self._logger.info(f"Training completed in {time_str}")

        metrics_mean_std = _compute_metrics_mean_std(metrics_full)
        return res, metrics_full, metrics_mean_std, trained_models


class GridSearchLearner:
    def __init__(
        self,
        model_types: dict,
        problem_type: ProblemTypes,
        data_split: DataSplitOptions,
        logger: Logger | None = None,
        n_cpus: int = _MAX_CPUS,
    ) -> None:
        self._logger = logger
        self._model_types: dict[str, Any] = model_types
        self._problem_type = problem_type
        self._data_split = data_split
        self._metrics = get_metrics(self._problem_type, logger=self._logger)
        self._n_cpus = n_cpus

        self._logger.info(f"Using {self._n_cpus} of {cpu_count()} CPUs for training")

    def fit(self, data: TabularData) -> tuple[dict, dict, dict, dict]:
        """Fit models to the data using Grid Search with cross validation. Evaluates them
        and returns metrics and statistics with the models.

        Args:
            data (TabularData): The data to fit the models with.

        Returns:
            res (dict): Dictionary containing model predictions for
            each bootstrap sample.
            metrics_full (dict): Dictionary containing metric values for
            each bootstrap sample.
            metrics_mean_std (dict): Dictionary containing average and
            standard deviation of metric values across bootstrap samples.
            trained_models (dict): Dictionary containing
            trained models for each model type.
        """
        self._logger.info("Fitting models using Grid Search...")

        # Extract the data and convert to numpy arrays
        X_train = data.X_train[0].to_numpy()
        X_test = data.X_test[0].to_numpy()
        y_train = data.y_train[0].to_numpy()
        y_test = data.y_test[0].to_numpy()

        # Make grid search compatible scorers
        metrics = (
            REGRESSION_METRICS
            if self._problem_type == ProblemTypes.Regression
            else CLASSIFICATION_METRICS
        )
        if (
            np.unique(y_train).shape[0] > 2
            and self._problem_type == ProblemTypes.Classification
        ):
            scorers = {}
            for metric_name, metric in metrics.items():
                if metric_name == Metrics.Accuracy:
                    scorers[metric_name] = make_scorer(metric)
                elif metric_name == Metrics.ROC_AUC:
                    scorers[metric_name] = make_scorer(
                        metric, multi_class="ovr", needs_proba=True
                    )
                else:
                    scorers[metric_name] = make_scorer(metric, average="micro")
            cv = StratifiedKFold(n_splits=self._data_split.k_folds, shuffle=True)
        else:
            scorers = {key: make_scorer(value) for key, value in metrics.items()}
            cv = self._data_split.k_folds

        # Fit models
        res = {0: {}}
        metrics_full = {}
        trained_models = {model_name: [] for model_name in self._model_types.keys()}
        metrics_mean_std = {model_name: {} for model_name in self._model_types.keys()}

        training_start = time()
        for model_name, params in self._model_types.items():
            res[0][model_name] = {}
            # Set up grid search
            refit = (
                "R2" if self._problem_type == ProblemTypes.Regression else "accuracy"
            )
            model_type = get_model_type(model_name, self._problem_type)
            model = get_model(model_type)

            gs = GridSearchCV(
                estimator=model,
                param_grid=params["params"],
                scoring=scorers,
                refit=refit,
                cv=cv,
                return_train_score=True,
                n_jobs=self._n_cpus,
            )

            # Fit the model
            self._logger.info(f"Fitting {model_name}...")
            gs.fit(X_train, y_train)

            # Make predictions for evaluation
            y_pred_train = gs.predict(X_train)
            res[0][model_name]["y_pred_train"] = y_pred_train
            y_pred_test = gs.predict(X_test)
            res[0][model_name]["y_pred_test"] = y_pred_test
            if self._problem_type == ProblemTypes.Classification:
                y_pred_probs_train = gs.predict_proba(X_train)
                res[0][model_name]["y_pred_train_proba"] = y_pred_probs_train
                y_pred_probs_test = gs.predict_proba(X_test)
                res[0][model_name]["y_pred_test_proba"] = y_pred_probs_test
            else:
                y_pred_probs_train = None
                y_pred_probs_test = None
            if model_name not in metrics_full:
                metrics_full[model_name] = []
            metrics_full[model_name].append(
                _evaluate(
                    model_name,
                    metrics,
                    y_train,
                    y_pred_train,
                    y_pred_probs_train,
                    y_test,
                    y_pred_test,
                    y_pred_probs_test,
                    self._logger,
                    self._problem_type,
                )
            )

            # append the best estimator
            trained_models[model_name].append(gs.best_estimator_)
            metrics_mean_std[model_name].update(
                self._compute_metrics_statistics(gs.cv_results_, gs.best_index_)
            )

        training_end = time()
        elapsed = training_end - training_start
        hours = int(elapsed) // 3600
        minutes = (int(elapsed) % 3600) // 60
        seconds = int(elapsed) % 60
        # Create format hh:mm:ss
        time_str = f"{hours:02d}:{minutes:02d}:{seconds:02d}"
        self._logger.info(f"Training completed in {time_str}")

        return res, metrics_full, metrics_mean_std, trained_models

    def _compute_metrics_statistics(self, cv_results: dict, best_index: int) -> dict:
        """
        Compute metric statistics for each model.

        Args:
            - cv_results (dict): The cross-validation results for a grid search.
            - best_index (int): The index of the best performing model.

        Returns:
            - dict: Dictionary containing metric statistics for
            the model.
        """
        metric_names = (
            REGRESSION_METRICS.keys()
            if self._problem_type == ProblemTypes.Regression
            else CLASSIFICATION_METRICS.keys()
        )

        statistics = {"train": {}, "test": {}}
        for metric in metric_names:
            statistics["train"][metric] = {
                "mean": cv_results[f"mean_train_{metric}"][best_index],
                "std": cv_results[f"std_train_{metric}"][best_index],
            }
            statistics["test"][metric] = {
                "mean": cv_results[f"mean_test_{metric}"][best_index],
                "std": cv_results[f"std_test_{metric}"][best_index],
            }

        return statistics


def _evaluate(
    model_name: str,
    metrics: dict,
    y_train: np.ndarray,
    y_pred_train: np.ndarray,
    y_pred_probs_train: np.ndarray,
    y_test: np.ndarray,
    y_pred_test: np.ndarray,
    y_pred_probs_test: np.ndarray,
    logger: object,
    problem_type: ProblemTypes,
) -> dict:
    """
    Evaluates the performance of a model using specified metrics.

    Args:
        - model_name (str): Name of the model being evaluated.
        - metrics (dict): The metrics to use in evaluation.
        - y_train (np.ndarray): True labels for the training set.
        - y_pred_train (np.ndarray): Predicted labels for the training set.
        - y_pred_probs_train (np.ndarray): Predicted probabilities for the training set.
        - y_test (np.ndarray): True labels for the test set.
        - y_pred_test (np.ndarray): Predicted labels for the test set.
        - y_pred_probs_test (np.ndarray): Predicted probabilities for the test set.
        - logger (object): The logger.
    """
    logger.info(f"Starting evaluation for {model_name}...")
    eval_res = {}

    if y_pred_probs_test is not None and y_pred_probs_test.shape[1] < 3:
        problem = ProblemTypes.BinaryClassification
    elif y_pred_probs_test is not None:
        problem = ProblemTypes.MultiClassification

    for metric_name, metric in metrics.items():
        eval_res[metric_name] = {}
        logger.info(f"Evaluating {model_name} on metric {metric_name} (train/test)...")

        # Regression
        if problem_type == ProblemTypes.Regression:
            metric_train = _calculate_regression_metrics(y_train, y_pred_train, metric)
            metric_test = _calculate_regression_metrics(y_test, y_pred_test, metric)

        # Binary classification
        elif problem_type == ProblemTypes.Classification:
            metric_train = _calculate_classification_metrics(
                y_true=y_train,
                y_pred=y_pred_train,
                y_pred_probs=y_pred_probs_train,
                metric_function=metric,
                metric_name=metric_name,
                problem_type=problem,
            )
            metric_test = _calculate_classification_metrics(
                y_true=y_test,
                y_pred=y_pred_test,
                y_pred_probs=y_pred_probs_test,
                metric_function=metric,
                metric_name=metric_name,
                problem_type=problem,
            )

        eval_res[metric_name]["train"] = {
            "value": metric_train,
        }
        eval_res[metric_name]["test"] = {
            "value": metric_test,
        }
    logger.info(f"Finished evaluation for {model_name}.")
    return eval_res


def _calculate_regression_metrics(y_true: np.array, y_pred: np.array, metric_function):
    """
    Calculate regression metrics for a given model.

    Args:
        - y_true (np.ndarray): True labels.
        - y_pred (np.ndarray): Predicted labels.
        - metric_function: Metric to calculate.

    Returns:
        - float: Value of the metric.
    """
    metric = metric_function(y_true, y_pred)

    return metric


def _calculate_classification_metrics(
    y_true: np.array,
    y_pred: np.array,
    y_pred_probs: np.array,
    metric_function,
    metric_name: Metrics,
    problem_type: ProblemTypes,
):
    """
    Calculate classification metrics for a given model.

    Args:
        - y_true (np.ndarray): True labels.
        - y_pred (np.ndarray): Predicted labels.
        - y_pred_probs (np.ndarray): Predicted probabilities.
        - metric_function: Metric to calculate.
        - metric_name (Metrics): Name of the metric.
        - problem_type (ProblemTypes): Type of classification problem.

    Returns:
        - float: Value of the metric.
    """

    if problem_type == ProblemTypes.BinaryClassification:
        if metric_name == Metrics.ROC_AUC:
            metric = metric_function(y_true, y_pred_probs[:, 1])
        else:
            metric = metric_function(y_true, y_pred)

    elif problem_type == ProblemTypes.MultiClassification:
        if metric_name == Metrics.Accuracy:
            metric = metric_function(y_true, y_pred)
        elif metric_name == Metrics.ROC_AUC:
            metric = metric_function(y_true, y_pred_probs, multi_class="ovr")
        else:
            metric = metric_function(y_true, y_pred, average="micro")

    return metric


def _compute_metrics_mean_std(metrics_full: dict) -> dict:
    """
    Go through the metric values for each model and calculate the mean
    and standard deviation of the metrics scores across all
    bootstraps/folds.

    Args:
        - metrics_full (dict): Dictionary containing metric values
        for each bootstrap sample.

    Returns:
        - dict: Dictionary containing metric statistics for
        each model.
    """
    statistics = {}

    for model_name, metrics_list in metrics_full.items():
        # Initialize dictionaries to store metric values for train and test sets
        train_metrics = {}
        test_metrics = {}

        # Aggregate metric values from each bootstrap sample
        for metrics in metrics_list:
            for metric_name, metric_values in metrics.items():
                if metric_name not in train_metrics:
                    train_metrics[metric_name] = []
                    test_metrics[metric_name] = []

                train_metrics[metric_name].append(metric_values["train"]["value"])
                test_metrics[metric_name].append(metric_values["test"]["value"])

        # Compute average and standard deviation for each metric
        statistics[model_name] = {"train": {}, "test": {}}
        for metric_name in train_metrics.keys():
            train_values = np.array(train_metrics[metric_name])
            test_values = np.array(test_metrics[metric_name])

            statistics[model_name]["train"][metric_name] = {
                "mean": np.mean(train_values),
                "std": np.std(train_values),
            }
            statistics[model_name]["test"][metric_name] = {
                "mean": np.mean(test_values),
                "std": np.std(test_values),
            }

    return statistics



from typing import Any

import pandas as pd
import shap
from sklearn.inspection import permutation_importance

from helix.utils.logging_utils import Logger


def calculate_global_shap_values(
    model,
    X: pd.DataFrame,
    logger: Logger,
) -> tuple[pd.DataFrame, Any]:
    """Calculate SHAP values for a given model and dataset.

    Args:
        model: Model object.
        X (pd.DataFrame): The dataset.
        logger (Logger): The logger.

    Returns:
        tuple[pd.DataFrame, Any]: SHAP dataframe and SHAP values.
    """
    logger.info(f"Calculating SHAP Importance for {model.__class__.__name__} model..")

    explainer = shap.Explainer(model.predict, X)

    shap_values = explainer(X)

    # Calculate Average Importance + set column names as index
    shap_df = (
        pd.DataFrame(shap_values.values, columns=X.columns).abs().mean().to_frame()
    )
    shap_df.columns = ["SHAP Importance"]

    logger.info("SHAP Importance Analysis Completed..")

    # Return the DataFrame
    return shap_df, shap_values


def calculate_permutation_importance(
    model,
    X: pd.DataFrame,
    y: pd.Series,
    permutation_importance_scoring: str,
    permutation_importance_repeat: int,
    random_state: int,
    logger: Logger,
):
    """Calculate permutation importance for a given model and dataset.

    Args:
        model: Model object.
        X (pd.DataFrame): Input features.
        y (pd.Series): Target variable.
        permutation_importance_scoring (str): Permutation importance scoring method.
        permutation_importance_repeat (int): Number of repeats for importance scoring.
        random_state (int): Seed for the random state.
        logger (Logger): The logger.

    Returns:
        permutation_importance: Permutation importance values
    """

    logger.info(
        f"Calculating Permutation Importance for {model.__class__.__name__} model.."
    )

    # Use permutation importance in sklearn.inspection to calculate feature importance
    permutation_importance_results = permutation_importance(
        model,
        X=X,
        y=y,
        scoring=permutation_importance_scoring,
        n_repeats=permutation_importance_repeat,
        random_state=random_state,
    )
    # Create a DataFrame with the results
    permutation_importance_df = pd.DataFrame(
        permutation_importance_results.importances_mean,
        index=X.columns,
        columns=["Permutation Importance"],
    )

    logger.info("Permutation Importance Analysis Completed..")

    # Return the DataFrame
    return permutation_importance_df



from dataclasses import dataclass

from helix.options.enums import ProblemTypes


@dataclass
class ExecutionOptions:
    experiment_name: str = "test"
    random_state: int = 1221
    problem_type: ProblemTypes = ProblemTypes.Classification
    dependent_variable: str | None = None



from pathlib import Path

import pandas as pd
import streamlit as st

from helix.components.configuration import (
    display_options,
    load_execution_options,
    load_plot_options,
)
from helix.components.experiments import experiment_selector
from helix.components.images.logos import sidebar_logo
from helix.components.logs import log_box
from helix.components.plot_editor import custom_plot_creator
from helix.components.plots import (
    display_metrics_table,
    display_predictions,
    fuzzy_feature_impacts,
    plot_box,
    plot_box_v2,
)
from helix.options.enums import (
    FeatureImportanceStateKeys,
    MachineLearningStateKeys,
    ViewExperimentKeys,
)
from helix.options.file_paths import (
    data_analysis_plots_dir,
    execution_options_path,
    fi_plot_dir,
    fuzzy_plot_dir,
    fuzzy_result_dir,
    helix_experiments_base_dir,
    log_dir,
    ml_metrics_mean_std_path,
    ml_plot_dir,
    ml_predictions_path,
    plot_options_path,
)
from helix.services.experiments import get_experiments
from helix.services.logs import get_logs

st.set_page_config(
    page_title="View Experiment",
    page_icon=sidebar_logo(),
)

header = st.session_state.get(ViewExperimentKeys.ExperimentName)

st.header(header if header is not None else "View Experiment")
st.write(
    """
    On this page, you can select one of your experiments to view.

    Use the dropdown below to see the details of your experiment.

    If you have not run any analyses yet, your experiment will be empty.
    Go to the sidebar on the **left** and select an analysis to run.
    """
)


choices = get_experiments()
experiment_name = experiment_selector(choices)


def display_experiment_plots(experiment_path: Path) -> None:
    """Display all available plots for the experiment.

    Args:
        experiment_path: Path to the experiment directory
    """
    # Data analysis plots
    data_analysis = data_analysis_plots_dir(experiment_path)
    if data_analysis.exists():
        plot_box(data_analysis, "Data analysis plots")

    # ML metrics and plots
    ml_metrics = ml_metrics_mean_std_path(experiment_path)
    if ml_metrics.exists():
        display_metrics_table(ml_metrics)

    ml_plots = ml_plot_dir(experiment_path)
    if ml_plots.exists():
        plot_box(ml_plots, "Machine learning plots")

    # Predictions
    predictions = ml_predictions_path(experiment_path)
    if predictions.exists():
        preds = pd.read_csv(predictions)
        display_predictions(preds)

    # Feature importance plots
    fi_plots = fi_plot_dir(experiment_path)
    if fi_plots.exists():
        mean_plots = [
            p
            for p in fi_plots.iterdir()
            if p.name.endswith("-all-folds-mean.png")  # mean global FI
            or p.name.startswith("local-")  # local plots
            or p.name.startswith("ensemble-")  # ensemble plots
        ]
        plot_box_v2(mean_plots, "Feature importance plots")

    # # Fuzzy plots
    fuzzy_plots = fuzzy_plot_dir(experiment_path)
    fuzzy_impacts = fuzzy_result_dir(experiment_path) / "top contextual rules.csv"
    if fuzzy_plots.exists():
        plot_box(fuzzy_plots, "Fuzzy plots")
    if fuzzy_impacts.exists():
        impacts_df = pd.read_csv(fuzzy_impacts)
        impacts_df.rename({"Unnamed: 0": ""}, axis=1, inplace=True)
        fuzzy_feature_impacts(impacts_df)


def display_experiment_logs(experiment_path: Path) -> None:
    """Display all available logs for the experiment.

    Args:
        experiment_path: Path to the experiment directory
    """
    log_configs = [
        ("ml", MachineLearningStateKeys.MLLogBox, "Machine learning logs"),
        ("fi", FeatureImportanceStateKeys.FILogBox, "Feature importance logs"),
        # ("fuzzy", FuzzyStateKeys.FuzzyLogBox, "Fuzzy FI Logs"),
    ]

    for log_dir_name, state_key, title in log_configs:
        try:
            st.session_state[state_key] = get_logs(
                log_dir(experiment_path) / log_dir_name
            )
            log_box(box_title=title, key=state_key)
        except NotADirectoryError:
            pass


def edit_results_plots(experiment_path: Path) -> None:

    predictions = ml_predictions_path(experiment_path)

    plot_opts_path = plot_options_path(experiment_path)

    execution_opts_path = execution_options_path(experiment_path)

    create_plot = st.checkbox(
        "Create custom results plot",
        key=ViewExperimentKeys.ShowCustomPlotCreator,
        help="Enable to create a custom plot for the results of your experiment.",
        value=False,
    )

    if (
        predictions.exists()
        and plot_opts_path.exists()
        and execution_opts_path.exists()
        and create_plot
    ):

        plot_opts = load_plot_options(plot_opts_path)
        execution_opts = load_execution_options(execution_opts_path)
        preds = pd.read_csv(predictions)

        plot = custom_plot_creator(
            predictions=preds,
            plot_opts=plot_opts,
            exec_opts=execution_opts,
        )

        if plot is not None:
            st.pyplot(plot, use_container_width=True)

            plot_name = st.text_input(
                "Enter a name for the plot (without extension):",
                value="custom_plot",
                key=ViewExperimentKeys.CustomPlotName,
                help="Name of the plot to save. Please provide any name you like, but make it as descriptive as possible. For example, 'RF vs XGBoost Test Set'",
            )

            if not plot_name.endswith(".png"):
                plot_name += ".png"

            if not plot_name.startswith("custom_plot"):
                plot_name = "custom plots-" + plot_name

            plot_path = ml_plot_dir(experiment_path) / plot_name

            save_plot = st.button("Save plot", key=ViewExperimentKeys.SavePlotButton)

            if plot_name and save_plot:
                plot.savefig(plot_path, bbox_inches="tight")
                st.success(f"Plot saved to {plot_path}")


if experiment_name:
    base_dir = helix_experiments_base_dir()
    experiment_path = base_dir / experiment_name
    display_options(experiment_path)
    display_experiment_plots(experiment_path)
    edit_results_plots(experiment_path)
    display_experiment_logs(experiment_path)



"""Component for editing plot appearance."""

import matplotlib.pyplot as plt
import pandas as pd
import streamlit as st

from helix.options.choices.ui import PLOT_FONT_FAMILIES
from helix.options.enums import PlotOptionKeys, PlotTypes, ProblemTypes
from helix.options.execution import ExecutionOptions
from helix.options.plotting import PlottingOptions
from helix.services.plotting import plot_scatter


def get_safe_index(value: str, options: list[str], default_value: str) -> int:
    """Safely get the index of a value in a list, returning the index of a default if not found.

    Args:
        value: The value to find in the list
        options: List of options to search in
        default_value: Default value to use if value is not found

    Returns:
        Index of the value in the list, or index of default_value
    """
    try:
        return options.index(value)
    except ValueError:
        return options.index(default_value)


def edit_plot_form(
    plot_opts: PlottingOptions, plot_type: PlotTypes, key_prefix: str = ""
):
    """
    Form to edit the appearance of plots.
    This form allows users to change the color scheme, font sizes, axis labels,
    and other visual aspects of the plots.

    Args:
        plot_opts (PlottingOptions): Current plotting options to edit.
        plot_type (PlotTypes): Type of the plot being edited.

    Returns:
        PlottingOptions: Updated plotting options based on user input.

    """

    with st.expander("Edit plot", expanded=False):

        st.subheader("Colours and styles")

        colour_scheme = st.selectbox(
            "Color scheme",
            plt.style.available,
            index=get_safe_index(
                plot_opts.plot_colour_scheme, plt.style.available, "default"
            ),
            key=key_prefix + PlotOptionKeys.ColourScheme,
            help="Select the color scheme for the plot",
        )

        # Color map for heatmaps
        if plot_type in [PlotTypes.CorrelationHeatmap, PlotTypes.TSNEPlot]:
            st.selectbox(
                "Color map",
                plt.colormaps(),
                index=get_safe_index(
                    plot_opts.plot_colour_map,
                    plt.colormaps(),
                    "viridis",
                ),
                key=key_prefix + PlotOptionKeys.ColourMap,
                help="Select the color map for heatmaps",
            )

        if plot_type in [PlotTypes.TargetVariableDistribution]:
            st.color_picker(
                "Plot colour",
                value="#1f77b4",
                key=key_prefix + PlotOptionKeys.PlotColour,
                help="Select the colour for the plot",
            )

        st.subheader("Customise plot titles")

        # Custom plot title
        plot_title = st.text_input(
            "Plot title",
            value=None,
            key=key_prefix + PlotOptionKeys.PlotTitle,
            help="Set a custom title for the plot",
        )

        col1, col2 = st.columns(2)
        with col1:
            yaxis_label = st.text_input(
                "Y-axis label",
                value=None,
                key=key_prefix + PlotOptionKeys.YAxisLabel,
                help="Set the label for the y-axis",
            )

        with col2:
            xaxis_label = st.text_input(
                "X-axis label",
                value=None,
                key=key_prefix + PlotOptionKeys.XAxisLabel,
                help="Set the label for the x-axis",
            )

        st.subheader("Font settings")

        col1, col2 = st.columns(2)

        with col1:
            title_font_size = st.number_input(
                "Title font size",
                min_value=8,
                max_value=50,
                value=plot_opts.plot_title_font_size,
                key=key_prefix + PlotOptionKeys.TitleFontSize,
                help="Set the font size for plot titles",
            )

            axis_font_size = st.number_input(
                "Axis font size",
                min_value=8,
                max_value=35,
                value=plot_opts.plot_axis_font_size,
                key=key_prefix + PlotOptionKeys.AxisFontSize,
                help="Set the font size for axis labels",
            )

            rotate_x = st.number_input(
                "X-axis label rotation",
                min_value=0,
                max_value=90,
                value=plot_opts.angle_rotate_xaxis_labels,
                key=key_prefix + PlotOptionKeys.RotateXAxisLabels,
                help="Set the rotation angle for x-axis labels",
            )

        with col2:

            tick_size = st.number_input(
                "Tick label size",
                min_value=6,
                max_value=35,
                value=plot_opts.plot_axis_tick_size,
                key=key_prefix + PlotOptionKeys.AxisTickSize,
                help="Set the font size for axis tick labels",
            )

            font_family = st.selectbox(
                "Font family",
                PLOT_FONT_FAMILIES,
                index=get_safe_index(
                    plot_opts.plot_font_family,
                    PLOT_FONT_FAMILIES,
                    "sans-serif",
                ),
                key=key_prefix + PlotOptionKeys.FontFamily,
                help="Select the font family for all text elements",
            )

            rotate_y = st.number_input(
                "Y-axis label rotation",
                min_value=0,
                max_value=90,
                value=plot_opts.angle_rotate_yaxis_labels,
                key=key_prefix + PlotOptionKeys.RotateYAxisLabels,
                help="Set the rotation angle for y-axis labels",
            )

        # Plot dimensions
        st.subheader("Plot dimensions")
        col1, col2 = st.columns(2)
        with col1:
            width = st.number_input(
                "Width (inches)",
                min_value=4,
                max_value=20,
                value=plot_opts.width,
                key=key_prefix + PlotOptionKeys.Width,
                help="Set the width of the plot in inches",
            )

        with col2:
            height = st.number_input(
                "Height (inches)",
                min_value=3,
                max_value=20,
                value=plot_opts.height,
                key=key_prefix + PlotOptionKeys.Height,
                help="Set the height of the plot in inches",
            )

        # Plot quality
        dpi = st.slider(
            "Image resolution (DPI)",
            min_value=330,
            max_value=2000,
            value=plot_opts.dpi if 330 < plot_opts.dpi <= 2000 else 330,
            key=key_prefix + PlotOptionKeys.DPI,
            help="Set the dots per inch (resolution) of the plot",
        )

    return PlottingOptions(
        plot_axis_font_size=axis_font_size,
        plot_axis_tick_size=tick_size,
        plot_colour_scheme=colour_scheme,
        dpi=dpi,
        angle_rotate_xaxis_labels=rotate_x,
        angle_rotate_yaxis_labels=rotate_y,
        save_plots=plot_opts.save_plots,
        plot_title_font_size=title_font_size,
        plot_font_family=font_family,
        height=height,
        width=width,
        plot_colour_map=st.session_state.get(
            key_prefix + PlotOptionKeys.ColourMap, plot_opts.plot_colour_map
        ),
        plot_title=plot_title,
        yaxis_label=yaxis_label,
        xaxis_label=xaxis_label,
        plot_colour=st.session_state.get(key_prefix + PlotOptionKeys.PlotColour, None),
    )


def custom_plot_creator(
    predictions: pd.DataFrame, plot_opts: PlottingOptions, exec_opts: ExecutionOptions
) -> None:

    style_by = [None]

    model_list = predictions["Model Name"].unique()

    col1, col2 = st.columns(2)

    with col1:

        models = st.multiselect(
            "Select models to plot",
            options=model_list,
            key=PlotOptionKeys.SelectedModels,
            help="Select the models you want to plot.",
        )

        if len(models) > 1:
            style_by.append("Model Name")

    with col2:
        set = st.multiselect(
            "Select the sets to plot",
            options=["Train", "Test"],
            key=PlotOptionKeys.SelectedSets,
            help="Select the sets you want to plot.",
        )

        if len(set) > 1:
            style_by.append("Set")

    if "Fold" in predictions.columns:
        fold_list = predictions["Fold"].unique()
        folds = st.multiselect(
            "Select folds to plot",
            options=fold_list,
            key=PlotOptionKeys.SelectedFolds,
            help="Select the folds you want to plot.",
        )

        if len(folds) > 1:
            style_by.append("Fold")

    if len(style_by) > 1:

        with col1:
            st.selectbox(
                "Color by",
                options=style_by,
                index=len(style_by) - 1,
                key=PlotOptionKeys.ColorBy,
                help="Select how to style the plot.",
            )
        with col2:
            st.selectbox(
                "Style by",
                options=style_by,
                index=len(style_by) - 1,
                key=PlotOptionKeys.StyleBy,
                help="Select how to style the plot.",
            )

    else:
        st.session_state[PlotOptionKeys.ColorBy] = None
        st.session_state[PlotOptionKeys.StyleBy] = None

    with col1:
        point_border_colour = st.color_picker(
            "Select a point border colour",
            value="#000000",
            key=PlotOptionKeys.PointColour + "Border",
            help="Select a border colour for the points in the scatter plot.",
        )

    if st.session_state.get(PlotOptionKeys.ColorBy, None) is None:
        with col2:
            st.color_picker(
                "Select a point colour",
                value="#1f77b4",
                key=PlotOptionKeys.PointColour,
                help="Select a colour for the points in the scatter plot.",
            )
            st.session_state[PlotOptionKeys.ColourMap + PlotTypes.ParityPlot.value] = (
                None
            )

    else:
        with col2:
            st.selectbox(
                "Select colour map for the points",
                options=[
                    "tab10",
                    "pastel",
                    "muted",
                    "bright",
                    "deep",
                    "colorblind",
                    "dark",
                ],
                index=0,
                key=PlotOptionKeys.ColourMap + PlotTypes.ParityPlot.value,
            )

    point_size = st.slider(
        "Point size",
        min_value=1,
        max_value=150,
        value=75,
        key=PlotOptionKeys.PointSize,
        help="Set the size of the points in the scatter plot.",
    )

    preds_filtered = predictions[
        predictions["Model Name"].isin(
            st.session_state.get(PlotOptionKeys.SelectedModels, [])
        )
    ]
    preds_filtered = preds_filtered[
        preds_filtered["Set"].isin(
            st.session_state.get(PlotOptionKeys.SelectedSets, [])
        )
    ]

    if "Fold" in preds_filtered.columns:
        preds_filtered = preds_filtered[
            preds_filtered["Fold"].isin(
                st.session_state.get(PlotOptionKeys.SelectedFolds, [])
            )
        ]

    hue = preds_filtered.get(st.session_state.get(PlotOptionKeys.ColorBy, None), None)
    style = preds_filtered.get(st.session_state.get(PlotOptionKeys.StyleBy, None), None)

    plot_settings = edit_plot_form(
        plot_opts,
        PlotTypes.ParityPlot,
    )

    if not preds_filtered.empty:
        if exec_opts.problem_type == ProblemTypes.Regression:
            scatter = plot_scatter(
                y=preds_filtered["Y True"],
                yp=preds_filtered["Y Prediction"],
                r2=None,
                dependent_variable=exec_opts.dependent_variable,
                plot_opts=plot_settings,
                point_colour=st.session_state.get(PlotOptionKeys.PointColour, None),
                edge_color=point_border_colour,
                point_size=point_size,
                show_grid=True,
                hue=hue,
                style_by=style,
                palette=st.session_state.get(
                    PlotOptionKeys.ColourMap + PlotTypes.ParityPlot.value,
                    None,
                ),
            )

            return scatter
        else:
            st.warning(
                "Custom plots are currently only supported for regression problems."
            )



from pathlib import Path


def uploaded_file_path(file_name: str, experiment_path: Path) -> Path:
    """Create the full upload path for data file uploads.

    Args:
        file_name (str): The name of the file.
        experiment_path (Path): The path of the experiment.

    Returns:
        Path: The full upload path for the file.
    """
    return experiment_path / file_name


def preprocessed_data_path(file_name: str, experiment_path: Path) -> Path:
    """Return the full path for the preprocessed version of a file.

    This function safely handles file names containing multiple dots.

    Args:
        file_name (str): The original file name, e.g. 'data.v1.csv'.
        experiment_path (Path): The experiment directory path.

    Returns:
        Path: The path to the preprocessed file inside the experiment directory.
    """
    file_path = Path(file_name)
    stem = file_path.stem  # 'data.v1'
    suffix = file_path.suffix  # '.csv'
    new_name = f"{stem}_preprocessed{suffix}"
    return experiment_path / new_name


def log_dir(experiment_path: Path) -> Path:
    """Create the full upload path for experiment log files.

    Args:
        experiment_path (str): The path of the experiment.

    Returns:
        Path: The full path for the log directory.
    """
    return experiment_path / "logs"


def ml_plot_dir(experiment_path: Path) -> Path:
    """Create the full path to the directory to save Machine Learning plots.

    Args:
        experiment_path (Path): The path of the experiment.

    Returns:
        Path: The full path for the Machine Learning plot directory.
    """
    return experiment_path / "plots" / "ml"


def ml_model_dir(experiment_path: Path) -> Path:
    """Create the full path to the directory to save Machine Learning models.

    Args:
        experiment_path (Path): The path of the experiment.

    Returns:
        Path: The full path for the Machine Learning model directory.
    """
    return experiment_path / "models"


def ml_metrics_mean_std_path(experiment_path: Path) -> Path:
    """Create the full path to the file to save the ML models' metrics'cmean
    and standard deviations.

    Args:
        experiment_path (Path): The path of the experiment.

    Returns:
        Path: The full path for the ml models metrics file.
    """
    return experiment_path / "results" / "ml_metrics" / "metrics_mean_std.json"


def ml_metrics_full_path(experiment_path: Path) -> Path:
    """Create the full path to the file to save the full ML models metrics results.

    Args:
        experiment_path (Path): The path of the experiment.

    Returns:
        Path: The full path for the ml models metrics file.
    """
    return experiment_path / "results" / "ml_metrics" / "metrics_full.json"


def ml_predictions_path(experiment_path: Path) -> Path:
    """Create the full path to the directory to save ML models predictions.

    Args:
        experiment_path (Path): The path of the experiment.

    Returns:
        Path: The full path for the ml models predictions file.
    """
    return experiment_path / "results" / "ml_metrics" / "predictions.csv"


def fi_plot_dir(experiment_path: Path) -> Path:
    """Create the full path to the directory to save Feature Importance plots.

    Args:
        experiment_path (Path): The path of the experiment.

    Returns:
        Path: The full path for the Feature Importance plot directory.
    """
    return experiment_path / "plots" / "fi"


def fi_result_dir(experiment_path: Path) -> Path:
    """Create the full path to the directory to save Feature Importance results.

    Args:
        experiment_path (Path): The path of the experiment.

    Returns:
        Path: The full path for the Feature Importance result directory.
    """
    return experiment_path / "results" / "fi"


def fi_options_dir(experiment_path: Path) -> Path:
    """Create the full path to the directory to save Feature Importance options.

    Args:
        experiment_path (Path): The path of the experiment.

    Returns:
        Path: The full path for the Feature Importance options directory.
    """
    return experiment_path / "options" / "fi"


def fuzzy_plot_dir(experiment_path: Path) -> Path:
    """Create the full path to the directory to save Fuzzy plots.

    Args:
        experiment_path (Path): The path of the experiment.

    Returns:
        Path: The full path for the Fuzzy plot directory.
    """
    return experiment_path / "plots" / "fuzzy"


def fuzzy_result_dir(experiment_path: Path) -> Path:
    """Create the full path to the directory to save Fuzzy results.

    Args:
        experiment_path (Path): The path of the experiment.

    Returns:
        Path: The full path for the Fuzzy result directory.
    """
    return experiment_path / "results" / "fuzzy"


def helix_experiments_base_dir() -> Path:
    """Return the path the base directory of all Helix experiments.

    This will be `/Users/<username>/HelixExperiments` on MacOS,
    `/home/<username>/HelixExperiments` on Linux, and
    `C:\\Users\\<username>\\HelixExperiments` on Windows.

    Returns:
        Path: The path to the Helix experiments base directory.
    """
    return Path.home() / "HelixExperiments"


def data_analysis_plots_dir(experiment_path: Path) -> Path:
    """Create the full path to the directory to save Data Analysis plots.

    Args:
        experiment_path (Path): The path of the experiment.

    Returns:
        Path: The full path for the Data Analysis plot directory.
    """
    return experiment_path / "plots" / "data_analysis"


def plot_options_path(experiment_path: Path) -> Path:
    """Return the path to an experiment's plot options.
    The path will be to a `json` file called `plot_options.json`

    Args:
        experiment_path (str): The path of the experiment.

    Returns:
        Path: The path to the experiment's plot options.

    Examples:
    ```python
    experiment_name = "test"
    experiment_path = helix_experiments_base_dir() / experiment_name
    plot_options_file = plot_options_path(experiment_path)
    ```
    """
    return experiment_path / "plot_options.json"


def execution_options_path(experiment_path: Path) -> Path:
    """Return the path to an experiment's execution options.
    The path will be to a `json` file called `execution_options.json`

    Args:
        experiment_path (str): The path of the experiment.

    Returns:
        Path: The path to the experiment's execution options.

    Examples:
    ```python
    experiment_name = "test"
    experiment_path = helix_experiments_base_dir() / experiment_name
    exec_options_file = execution_options_path(experiment_path)
    ```
    """
    return experiment_path / "execution_options.json"


def data_preprocessing_options_path(experiment_path: Path) -> Path:
    """Return the path to an experiment's data preprocessing options.
    The path will be to a `json` file called `data_preprocessing_options.json`

    Args:
        experiment_path (str): The path of the experiment.

    Returns:
        Path: The path to the experiment's data preprocessing options.

    Examples:
    ```python
    experiment_name = "test"
    experiment_path = helix_experiments_base_dir() / experiment_name
    data_preprocessing_options_file = data_preprocessing_options_path(experiment_path)
    ```
    """
    return experiment_path / "data_preprocessing_options.json"


def ml_options_path(experiment_path: Path) -> Path:
    """Return the path to an experiment's ML options.
    The path will be to a `json` file called `ml_options.json`

    Args:
        experiment_path (str): The path of the experiment.

    Returns:
        Path: The path to the experiment's ML options.

    Examples:
    ```python
    experiment_name = "test"
    experiment_path = helix_experiments_base_dir() / experiment_name
    ml_options_file = ml_options_path(experiment_path)
    ```
    """
    return experiment_path / "ml_options.json"


def fi_options_path(experiment_path: Path) -> Path:
    """Return the path to an experiment's FI options.
    The path will be to a `json` file called `fi_options.json`

    Args:
        experiment_path (str): The path of the experiment.

    Returns:
        Path: The path to the experiment's FI options.

    Examples:
    ```python
    experiment_name = "test"
    experiment_path = helix_experiments_base_dir() / experiment_name
    fi_options_file = fi_options_path(experiment_path)
    ```
    """
    return experiment_path / "fi_options.json"


def fuzzy_options_path(experiment_path: Path) -> Path:
    """Return the path to an experiment's Fuzzy options.
    The path will be to a `json` file called `fuzzy_options.json`

    Args:
        experiment_path (str): The path of the experiment.

    Returns:
        Path: The path to the experiment's Fuzzy options.

    Examples:
    ```python
    experiment_name = "test"
    experiment_path = helix_experiments_base_dir() / experiment_name
    fuzzy_options_file = fuzzy_options_path(experiment_path)
    ```
    """
    return experiment_path / "fuzzy_options.json"


def data_options_path(experiment_path: Path) -> Path:
    """Return the path to an experiment's data options.
    The path will be to a `json` file called `data_options.json`

    Args:
        experiment_path (str): The path of the experiment.

    Returns:
        Path: The path to the experiment's data options.

    Examples:
    ```python
    experiment_name = "test"
    experiment_path = helix_experiments_base_dir() / experiment_name
    data_options_file = data_options_path(experiment_path)
    ```
    """
    return experiment_path / "data_options.json"



from pathlib import Path
from typing import List, Tuple

import matplotlib.pyplot as plt
import pandas as pd
import streamlit as st
from sklearn.preprocessing import MinMaxScaler, StandardScaler

from helix.components.plot_editor import edit_plot_form
from helix.options.enums import DataAnalysisStateKeys, Normalisations, PlotTypes
from helix.options.plotting import PlottingOptions
from helix.services.plotting import (
    create_pairplot,
    create_tsne_plot,
    plot_correlation_heatmap,
    plot_target_variable_distribution,
)


@st.experimental_fragment
def target_variable_dist_form(
    data,
    dep_var_name,
    data_analysis_plot_dir,
    plot_opts: PlottingOptions,
    key_prefix: str = "",
):
    """
    Form to create the target variable distribution plot.

    Uses plot-specific settings that are not saved between sessions.
    """

    show_kde = st.toggle(
        "Show kernel density estimation in the distribution plot",
        value=True,
        key=f"{key_prefix}_{DataAnalysisStateKeys.ShowKDE}",
    )
    n_bins = st.slider(
        "Number of bins",
        min_value=5,
        max_value=50,
        value=10,
        key=f"{key_prefix}_{DataAnalysisStateKeys.NBins}",
    )

    plot_settings = edit_plot_form(
        plot_opts,
        plot_type=PlotTypes.TargetVariableDistribution,
        key_prefix=key_prefix + PlotTypes.TargetVariableDistribution.value,
    )

    show_plot = st.checkbox(
        "Create target variable distribution plot",
        key=f"{key_prefix}_{DataAnalysisStateKeys.TargetVarDistribution}",
    )
    if show_plot:

        displot = plot_target_variable_distribution(
            data,
            show_kde,
            n_bins,
            plot_settings,
            dep_var_name,
        )

        st.pyplot(displot)
        plt.close()

        if st.button(
            "Save Plot",
            key=f"{key_prefix}_{DataAnalysisStateKeys.SaveTargetVarDistribution}",
        ):
            displot.savefig(
                data_analysis_plot_dir / f"{dep_var_name}_distribution_{key_prefix}.png"
            )
            plt.clf()
            st.success("Plot created and saved successfully.")


def _render_corr_block(
    *,
    corr: pd.DataFrame,
    plot_opts: PlottingOptions,
    key_prefix: str,
    mode: str,  # "basic" or "advanced"
    data_analysis_plot_dir: Path,
) -> None:
    """Shared UI for plot+data + save. Keeps keys/filenames stable."""
    plot_tab, df_tab = st.tabs(
        [
            f"{mode.capitalize()} Correlation Plot",
            f"{mode.capitalize()} Correlation Data",
        ]
    )

    with plot_tab:
        plot_settings = edit_plot_form(
            plot_opts,
            PlotTypes.CorrelationHeatmap,
            key_prefix + PlotTypes.CorrelationHeatmap.value + f"_{mode}",
        )

        show_plot = st.checkbox(
            f"Create {mode.capitalize()} Correlation Heatmap Plot",
            key=f"{key_prefix}_{DataAnalysisStateKeys.CorrelationHeatmap}_{mode}",
        )

        if show_plot:
            fig = plot_correlation_heatmap(corr, plot_settings)
            st.pyplot(fig)

            if st.button(
                f"Save {mode.capitalize()} Plot",
                key=f"{key_prefix}_{DataAnalysisStateKeys.SaveHeatmap}_{mode}",
            ):
                fig.savefig(
                    data_analysis_plot_dir
                    / f"correlation_heatmap_{key_prefix}_{mode}.png"
                )
                plt.clf()
                st.success(f"{mode.capitalize()} plot created and saved successfully.")

    with df_tab:
        st.dataframe(corr)


def _advanced_selectors(
    data: pd.DataFrame, key_prefix: str
) -> Tuple[List[str], List[str], bool]:
    """Row/column selectors for advanced mode; returns (rows, cols, enabled)."""
    variables_1, variables_2 = st.columns(2)

    with variables_1:
        st.markdown("###### Variables in the rows")

        default_rows = (
            list(data.columns)
            if st.toggle(
                "Select all variables (rows)",
                value=False,
                key=f"{key_prefix}_{DataAnalysisStateKeys.SelectAllDescriptorsCorrelation}_1",
            )
            else []
        )

        corr_rows = st.multiselect(
            "Select variables to include in the correlation rows",
            data.columns,
            default=default_rows,
            key=f"{key_prefix}_{DataAnalysisStateKeys.DescriptorCorrelation}_1",
        )

        if len(corr_rows) < 1:
            st.warning("Please select at least one row variable.")

    with variables_2:
        st.markdown("###### Variables in the columns")

        default_cols = (
            list(data.columns)
            if st.toggle(
                "Select all variables (columns)",
                value=False,
                key=f"{key_prefix}_{DataAnalysisStateKeys.SelectAllDescriptorsCorrelation}_2",
            )
            else []
        )

        corr_cols = st.multiselect(
            "Select variables to include in the correlation columns",
            data.columns,
            default=default_cols,
            key=f"{key_prefix}_{DataAnalysisStateKeys.DescriptorCorrelation}_2",
        )

        if len(corr_cols) < 1:
            st.warning("Please select at least one column variable.")

    enable_adv = (len(corr_rows) > 0) and (len(corr_cols) > 0)
    return corr_rows, corr_cols, enable_adv


@st.experimental_fragment
def correlation_heatmap_form(
    data, data_analysis_plot_dir, plot_opts: PlottingOptions, key_prefix: str = ""
):
    """
    - Always shows the basic (square) correlation first.
    - Advanced options are displayed at the end (below basic).
    - User can generate BOTH basic and advanced plots/data in the same run.
    """

    # ----------------------------
    # BASIC (non-advanced) section
    # ----------------------------
    st.markdown("###### Select the variables to show in the correlation plot")

    default_basic = (
        list(data.columns)
        if st.toggle(
            "Select all independent variables",
            value=False,
            key=f"{key_prefix}_{DataAnalysisStateKeys.SelectAllDescriptorsCorrelation}",
        )
        else []
    )

    corr_basic_vars = st.multiselect(
        "Select independent variables to include in the correlation heatmap",
        data.columns,
        default=default_basic,
        key=f"{key_prefix}_{DataAnalysisStateKeys.DescriptorCorrelation}",
    )

    if len(corr_basic_vars) < 2:
        st.warning(
            "Please select at least two variables to create the correlation heatmap."
        )

    enable_basic = len(corr_basic_vars) > 1

    calc_basic = st.checkbox(
        "Calculate Correlation Matrix (basic)",
        key=f"{key_prefix}_{DataAnalysisStateKeys.CalculateCorrelationMatrix}_basic",
        disabled=not enable_basic,
        value=enable_basic,
    )

    if calc_basic:
        corr_basic = data[corr_basic_vars].corr()
        _render_corr_block(
            corr=corr_basic,
            plot_opts=plot_opts,
            key_prefix=key_prefix,
            mode="basic",
            data_analysis_plot_dir=data_analysis_plot_dir,
        )

    st.divider()

    # ----------------------------
    # ADVANCED section (shown at end)
    # ----------------------------
    show_adv = st.checkbox(
        "Show advanced correlation heatmap options (selected variable to selected variable)",
        key=f"{key_prefix}_{DataAnalysisStateKeys.AdvancedCorrOptions}",
    )
    if not show_adv:
        return

    st.markdown(
        """
**What advanced mode does**
- Lets you choose **different variables for rows vs columns**.
- Produces a **rectangular correlation matrix** (rows × columns), not necessarily symmetric.
- Useful when you want to correlate one feature set against another subset.
        """
    )

    corr_rows, corr_cols, enable_adv = _advanced_selectors(data, key_prefix=key_prefix)

    calc_adv = st.checkbox(
        "Calculate Correlation Matrix (advanced)",
        key=f"{key_prefix}_{DataAnalysisStateKeys.CalculateCorrelationMatrix}_advanced",
        disabled=not enable_adv,
        value=enable_adv,
    )

    if calc_adv:
        corr_union = list(set(corr_rows) | set(corr_cols))
        corr_adv_full = data[corr_union].corr()
        corr_adv = corr_adv_full.loc[corr_rows, corr_cols]

        _render_corr_block(
            corr=corr_adv,
            plot_opts=plot_opts,
            key_prefix=key_prefix,
            mode="advanced",
            data_analysis_plot_dir=data_analysis_plot_dir,
        )


@st.experimental_fragment
def pairplot_form(  # noqa: C901
    data, data_analysis_plot_dir, plot_opts: PlottingOptions, key_prefix: str = ""
):
    """
    Form to create the pairplot plot.

    Uses plot-specific settings that are not saved between sessions.
    """

    if st.toggle(
        "Select all independent variables",
        value=False,
        key=f"{key_prefix}_{DataAnalysisStateKeys.SelectAllDescriptorsPairPlot}",
    ):
        default_corr = list(data.columns[:-1])
    else:
        default_corr = None

    descriptors = st.multiselect(
        "Select independent variables to include in the pairplot",
        data.columns[:-1],
        default=default_corr,
        key=f"{key_prefix}_{DataAnalysisStateKeys.DescriptorPairPlot}",
    )

    pairplot_data = data[descriptors + [data.columns[-1]]]

    if len(descriptors) < 1:
        st.warning(
            "Please select at least one independent variable to create the correlation plot."
        )

    exclude_corner = st.checkbox(
        "Exclude corner plot",
        value=True,
        key=f"{key_prefix}_{DataAnalysisStateKeys.ExcludeCorner}",
    )

    col1, col2 = st.columns(2)
    with col1:
        kind = st.selectbox(
            "Select how to display the correlation between variables",
            options=["scatter", "kde", "hist", "reg"],
            index=0,
            key=f"{key_prefix}_{DataAnalysisStateKeys.PairPlotCorrelationType}",
            help="Select the type of distribution to plot in the pairplot.",
        )
    with col2:
        diag_kind = st.selectbox(
            "Select the diagonal plot type",
            options=["hist", "kde", None],
            index=0,
            key=f"{key_prefix}_{DataAnalysisStateKeys.PairPlotDiagonalType}",
            help="Select the type of plot to use on the diagonal of the pairplot.",
        )

    plot_settings = edit_plot_form(
        plot_opts, PlotTypes.PairPlot, key_prefix + PlotTypes.PairPlot.value
    )

    show_plot = st.checkbox(
        "Create Pairplot", key=f"{key_prefix}_{DataAnalysisStateKeys.PairPlot}"
    )
    if show_plot:

        pairplot = create_pairplot(
            pairplot_data, plot_settings, exclude_corner, kind, diag_kind
        )

        st.pyplot(pairplot)
        plt.close()

        if st.button(
            "Save Plot", key=f"{key_prefix}_{DataAnalysisStateKeys.SavePairPlot}"
        ):
            pairplot.savefig(data_analysis_plot_dir / f"pairplot_{key_prefix}.png")
            plt.clf()
            st.success("Plot created and saved successfully.")


@st.experimental_fragment
def tSNE_plot_form(  # noqa: C901
    data,
    random_state,
    data_analysis_plot_dir,
    plot_opts: PlottingOptions,
    scaler: Normalisations = None,
    key_prefix: str = "",
):

    X = data.drop(columns=[data.columns[-1]])
    y = data[data.columns[-1]]

    if scaler == Normalisations.NoNormalisation:
        scaler = st.selectbox(
            "Select normalisation for comparison (this will not affect the normalisation for ML models)",
            options=[Normalisations.Standardisation, Normalisations.MinMax],
            key=f"{key_prefix}_{DataAnalysisStateKeys.SelectNormTsne}",
        )

    if scaler == Normalisations.MinMax:
        X_scaled = MinMaxScaler().fit_transform(X)
    elif scaler == Normalisations.Standardisation:
        X_scaled = StandardScaler().fit_transform(X)

    perplexity = st.slider(
        "Perplexity",
        min_value=5,
        max_value=50 if 50 < len(y) else len(y) - 1,
        value=30 if 50 < len(y) else int(len(y) / 2),
        help="The perplexity parameter controls the balance between local and global aspects of the data.",
        key=f"{key_prefix}_{DataAnalysisStateKeys.Perplexity}",
    )

    plot_settings = edit_plot_form(
        plot_opts, PlotTypes.TSNEPlot, key_prefix + PlotTypes.TSNEPlot.value
    )

    show_plot = st.checkbox(
        "Create t-SNE Plot", key=f"{key_prefix}_{DataAnalysisStateKeys.TSNEPlot}"
    )
    if show_plot:

        tsne_plot = create_tsne_plot(
            X,
            X_scaled,
            y,
            plot_settings,
            random_state,
            perplexity,
        )

        st.pyplot(tsne_plot)
        plt.close()

        if st.button(
            "Save Plot", key=f"{key_prefix}_{DataAnalysisStateKeys.SaveTSNEPlot}"
        ):
            tsne_plot.savefig(data_analysis_plot_dir / f"tsne_plot_{key_prefix}.png")
            plt.clf()
            st.success("Plots created and saved successfully.")



"""Feature importance form components for Helix.

This module contains form components for configuring feature importance analysis:
- Global feature importance methods (Permutation, SHAP)
- Ensemble methods (Mean, Majority Vote)
- Local feature importance methods (LIME, SHAP)
- Fuzzy feature importance options
"""

import streamlit as st

from helix.options.enums import (
    ExecutionStateKeys,
    FeatureImportanceStateKeys,
    FeatureImportanceTypes,
    FuzzyStateKeys,
    ProblemTypes,
)


@st.experimental_fragment
def fi_options_form():
    """Form for configuring feature importance options."""
    global_methods = {}

    st.write("### Global feature importance methods")
    st.write(
        "Select global methods to assess feature importance across the entire dataset. "
        "These methods help in understanding overall feature impact."
    )

    use_permutation = st.checkbox(
        "Permutation importance",
        help="Evaluate feature importance by permuting feature values.",
    )

    global_methods[FeatureImportanceTypes.PermutationImportance.value] = {
        "type": "global",
        "value": use_permutation,
    }

    use_shap = st.checkbox(
        "SHAP",
        help="Apply SHAP (SHapley Additive exPlanations) for global interpretability.",
    )
    global_methods[FeatureImportanceTypes.SHAP.value] = {
        "type": "global",
        "value": use_shap,
    }

    st.session_state[FeatureImportanceStateKeys.GlobalFeatureImportanceMethods] = (
        global_methods
    )

    st.write("### Ensemble feature importance methods")
    st.write(
        "Ensemble methods combine results from multiple feature importance techniques, "
        "enhancing robustness. Choose how to aggregate feature importance insights."
    )

    # global methods need to be set to perform ensemble methods
    ensemble_is_disabled = not (use_permutation or use_shap)
    if ensemble_is_disabled:
        st.warning(
            "You must configure at least one global feature importance method to perform ensemble methods.",
            icon="⚠",
        )
    ensemble_methods = {}
    use_mean = st.checkbox(
        "Mean",
        help="Calculate the mean importance score across methods.",
        disabled=ensemble_is_disabled,
    )
    ensemble_methods[FeatureImportanceTypes.Mean.value] = use_mean
    use_majority = st.checkbox(
        "Majority vote",
        help="Use majority voting to identify important features.",
        disabled=ensemble_is_disabled,
    )
    ensemble_methods[FeatureImportanceTypes.MajorityVote.value] = use_majority

    st.session_state[FeatureImportanceStateKeys.EnsembleMethods] = ensemble_methods

    st.write("### Local feature importance methods")
    st.write(
        "Select local methods to interpret individual predictions. "
        "These methods focus on explaining predictions at a finer granularity."
    )

    local_importance_methods = {}
    use_lime = st.checkbox(
        "LIME",
        help="Use LIME (Local Interpretable Model-Agnostic Explanations) for local interpretability.",
    )
    local_importance_methods[FeatureImportanceTypes.LIME.value] = {
        "type": "local",
        "value": use_lime,
    }
    use_local_shap = st.checkbox(
        "Local SHAP",
        help="Use SHAP for local feature importance at the instance level.",
    )
    local_importance_methods[FeatureImportanceTypes.SHAP.value] = {
        "type": "local",
        "value": use_local_shap,
    }

    st.session_state[FeatureImportanceStateKeys.LocalImportanceFeatures] = (
        local_importance_methods
    )

    st.write("### Additional configuration options")

    # Number of important features
    st.number_input(
        "Number of most important features to plot",
        min_value=1,
        value=10,
        help="Select how many top features to visualise based on their importance score.",
        key=FeatureImportanceStateKeys.NumberOfImportantFeatures,
    )

    # Scoring function for permutation importance
    if (
        st.session_state.get(ExecutionStateKeys.ProblemType, ProblemTypes.Auto).lower()
        == ProblemTypes.Regression
    ):
        scoring_options = [
            "neg_mean_absolute_error",
            "neg_root_mean_squared_error",
        ]
    elif (
        st.session_state.get(ExecutionStateKeys.ProblemType, ProblemTypes.Auto).lower()
        == ProblemTypes.Classification
    ):
        scoring_options = ["accuracy", "f1"]
    else:
        scoring_options = []

    st.selectbox(
        "Scoring function for permutation importance",
        scoring_options,
        help="Choose a scoring function to evaluate the model during permutation importance.",
        key=FeatureImportanceStateKeys.ScoringFunction,
    )

    # Number of repetitions for permutation importance
    st.number_input(
        "Number of repetitions for permutation importance",
        min_value=1,
        value=5,
        help="Specify the number of times to shuffle each feature for importance estimation.",
        key=FeatureImportanceStateKeys.NumberOfRepetitions,
    )

    # Fuzzy Options
    st.write("### Fuzzy Feature Importance Options")
    st.write(
        "Activate fuzzy methods to capture interactions between features in a fuzzy rule-based system. "
        "Define the number of features, clusters, and granular options for enhanced interpretability."
    )

    # both ensemble_methods and local_importance_methods
    fuzzy_is_disabled = (not (use_lime or use_local_shap)) or (
        not (use_mean or use_majority)
    )
    if fuzzy_is_disabled:
        st.warning(
            "You must configure both ensemble and local importance methods to use fuzzy feature selection.",
            icon="⚠",
        )
    fuzzy_feature_importance = st.checkbox(
        "Enable Fuzzy Feature Importance",
        help="Toggle fuzzy feature importance to analyze feature interactions.",
        key=FuzzyStateKeys.FuzzyFeatureSelection,
        disabled=fuzzy_is_disabled,
    )

    if fuzzy_feature_importance:
        st.number_input(
            "Number of features for fuzzy interpretation",
            min_value=1,
            value=5,
            help="Set the number of features for fuzzy analysis.",
            key=FuzzyStateKeys.NumberOfFuzzyFeatures,
        )

        st.checkbox(
            "Granular features",
            help="Divide features into granular categories for in-depth analysis.",
            key=FuzzyStateKeys.GranularFeatures,
        )

        st.number_input(
            "Number of clusters for target variable",
            min_value=2,
            value=5,
            help="Set the number of clusters to categorise the target variable for fuzzy interpretation.",
            key=FuzzyStateKeys.NumberOfClusters,
        )

        st.text_input(
            "Names of clusters (comma-separated)",
            help="Specify names for each cluster (e.g., Low, Medium, High).",
            key=FuzzyStateKeys.ClusterNames,
            value=", ".join(["very low", "low", "medium", "high", "very high"]),
        )

        st.number_input(
            "Number of top occurring rules for fuzzy synergy analysis",
            min_value=1,
            value=10,
            help="Set the number of most frequent fuzzy rules for synergy analysis.",
            key=FuzzyStateKeys.NumberOfTopRules,
        )

    st.subheader("Select outputs to save")

    # Save options
    st.toggle(
        "Save feature importance options",
        help="Save the selected configuration of feature importance methods.",
        key=FeatureImportanceStateKeys.SaveFeatureImportanceOptions,
        value=True,
    )

    st.toggle(
        "Save feature importance results",
        help="Store the results from feature importance computations.",
        key=FeatureImportanceStateKeys.SaveFeatureImportanceResults,
        value=True,
    )



"""Feature selection utilities for Helix."""

import numpy as np
import pandas as pd
from sklearn.feature_selection import VarianceThreshold


def sort_by_feature_name(df: pd.DataFrame) -> pd.DataFrame:
    """Sort features by name length to ensure consistent ordering.

    Args:
        df : pd.DataFrame
            Input DataFrame with features as columns

    Returns:
        pd.DataFrame
            DataFrame with columns sorted by name length
    """
    df = df.T
    df["len"] = df.T.columns.str.len()
    df_sorted = df.sort_values(["len"])
    df_sorted = df_sorted.drop(["len"], axis=1)
    return df_sorted.T


def remove_correlation(dataset: pd.DataFrame, threshold: float) -> pd.DataFrame:
    """Remove highly correlated features based on Pearson correlation.

    Args:

    dataset : pd.DataFrame
        Input DataFrame with features as columns
    threshold : float
        Correlation threshold. Features with correlation above this will be removed.

    Returns:

    pd.DataFrame
        DataFrame with correlated features removed
    """
    col_corr = set()  # Set of all the names of deleted columns
    corr_matrix = dataset.corr().abs()
    for i in range(len(corr_matrix.columns)):
        for j in range(i):
            if corr_matrix.iloc[i, j] >= threshold:
                colname = corr_matrix.columns[i]  # getting the name of column
                col_corr.add(colname)
                if colname in dataset.columns:
                    dataset.drop(
                        colname, axis=1, inplace=True
                    )  # deleting the column from the dataset
    return dataset


def remove_low_variance(X: pd.DataFrame, threshold: float = 0.01) -> pd.DataFrame:
    """Remove features with low variance.

    Args:
    X : pd.DataFrame
        Input DataFrame with features as columns
    threshold : float, default=0.01
        Features with variance below this threshold will be removed

    Returns:
    pd.DataFrame
        DataFrame with low variance features removed
    """
    selector = VarianceThreshold(threshold=threshold)
    selector.fit(X)
    return X.iloc[:, selector.get_support()]


def standard_error_prediction(y_true, y_pred) -> float:
    """Calculate Standard Error of Prediction.

    Args:
    y_true : array-like
        True target values
    y_pred : array-like
        Predicted target values

    Returns:
    float
        Standard error of prediction
    """
    return np.sqrt(np.mean((y_true - y_pred) ** 2))



from sklearn.ensemble import RandomForestClassifier as SkLearnRandomForestClassifier
from sklearn.ensemble import RandomForestRegressor as SkLearnRandomForestRegressor


class RandomForestClassifier(SkLearnRandomForestClassifier):
    """A Helix implementation of scikit-learn's RandomForestClassifier.

    It is exactly the same, except it always sets `n_jobs` to 1
    so that when it runs in a parallel context (e.g. Grid Search),
    it doesn't slow down the training by consuming too much CPU
    """

    def __init__(
        self,
        n_estimators=100,
        *,
        criterion="gini",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0,
        max_features="sqrt",
        max_leaf_nodes=None,
        min_impurity_decrease=0,
        bootstrap=True,
        oob_score=False,
        n_jobs=1,
        random_state=None,
        verbose=0,
        warm_start=False,
        class_weight=None,
        ccp_alpha=0,
        max_samples=None,
    ):
        super().__init__(
            n_estimators,
            criterion=criterion,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_features=max_features,
            max_leaf_nodes=max_leaf_nodes,
            min_impurity_decrease=min_impurity_decrease,
            bootstrap=bootstrap,
            oob_score=oob_score,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            warm_start=warm_start,
            class_weight=class_weight,
            ccp_alpha=ccp_alpha,
            max_samples=max_samples,
        )


class RandomForestRegressor(SkLearnRandomForestRegressor):
    """A Helix implementation of scikit-learn's RandomForestRegressor.

    It is exactly the same, except it always sets `n_jobs` to 1
    so that when it runs in a parallel context (e.g. Grid Search),
    it doesn't slow down the training by consuming too much CPU
    """

    def __init__(
        self,
        n_estimators=100,
        *,
        criterion="squared_error",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0,
        max_features=1,
        max_leaf_nodes=None,
        min_impurity_decrease=0,
        bootstrap=True,
        oob_score=False,
        n_jobs=1,
        random_state=None,
        verbose=0,
        warm_start=False,
        ccp_alpha=0,
        max_samples=None,
    ):
        super().__init__(
            n_estimators,
            criterion=criterion,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_features=max_features,
            max_leaf_nodes=max_leaf_nodes,
            min_impurity_decrease=min_impurity_decrease,
            bootstrap=bootstrap,
            oob_score=oob_score,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            warm_start=warm_start,
            ccp_alpha=ccp_alpha,
            max_samples=max_samples,
        )
